# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from transformers.trainer import TrainerCallback

from logix import LogIX, LogIXScheduler
from logix.huggingface.arguments import LogIXArguments
from logix.utils import merge_logs


class LogIXCallback(TrainerCallback):
    def __init__(
        self,
        logix: LogIX,
        logix_scheduler: LogIXScheduler,
        args: LogIXArguments,
    ):
        self.logix = logix
        self.logix_scheduler = iter(logix_scheduler)
        self.args = args

        self.accumulated_log = []

        self._log_dataloader = None

    def on_init_end(self, args, state, control, **kwargs):
        model = kwargs["model"]
        self.logix.watch(
            model=model,
            name_filter=self.args.name_filter,
            type_filter=self.args.type_filter,
        )

        if self.args.lora:
            self.logix.add_lora()

    def on_epoch_begin(self, args, state, control, **kwargs):
        if self.args.mode == "log":
            next(self.logix_scheduler)

    def on_epoch_end(self, args, state, control, **kwargs):
        if self.args.mode == "log":
            self.logix.finalize()

    def on_train_begin(self, args, state, control, **kwargs):
        if self.args.initialize_from_log:
            self.logix.initialize_from_log()

        if self.args.mode in ["influence", "self_influence"]:
            self.logix.setup({"grad": ["log"]})
            self.logix.eval()

            state.epoch = 0
            state.num_train_epochs = 1

    def on_step_end(self, args, state, control, **kwargs):
        if self.args.mode == "influence":
            self.accumulated_log.append(self.logix.get_log(copy=True))
            accumulated_log = merge_logs(self.accumulated_log)

            self.logix.influence.compute_influence_all(
                accumulated_log,
                self.log_dataloader(),
                mode=self.args.influence_mode,
                damping=self.args.influence_damping,
                influence_groups=self.args.influence_groups,
                save=True,
            )

            self.accumulated_log = []
        elif self.args.mode == "self_influence":
            self.accumulated_log.append(self.logix.get_log(copy=True))
            accumulated_log = merge_logs(self.accumulated_log)

            self.logix.influence.compute_self_influence(
                accumulated_log,
                damping=self.args.influence_damping,
                influence_groups=self.args.influence_groups,
            )

            self.accumulated_log = []

    def on_substep_end(self, args, state, control, **kwargs):
        if self.args.mode in ["influence", "self_influence"]:
            self.accumulated_log.append(self.logix.get_log(copy=True))

    def log_dataloader(self):
        if self._log_dataloader is None:
            self._log_dataloader = self.logix.build_log_dataloader(
                num_workers=self.args.log_num_workers,
                batch_size=self.args.log_batch_size,
            )
        return self._log_dataloader



# Copyright 2023-present the LogIX team and Huggingface PEFT team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import torch


def is_lora(model):
    """
    Check if LoRA is enabled for the model.
    """
    return any("logix_lora_B" in name for name, module in model.named_modules())


def compute_top_k_singular_vectors(matrix, k):
    """
    Compute the top k singular vectors of a matrix.
    """
    U, S, _ = torch.linalg.svd(matrix)
    top_k_singular_vectors = U[:, :k]
    top_k_singular_values = S[:k]
    return top_k_singular_vectors, top_k_singular_values


def find_parameter_sharing_group(
    module_name: str, parameter_sharing_groups: List[str] = None
):
    if parameter_sharing_groups is None:
        return "logix_lora_none"

    found_groups = [psg for psg in parameter_sharing_groups if psg in module_name]
    assert (
        len(found_groups) == 1
    ), "Each module can belong to only one parameter sharing group."
    return found_groups[0]


def _get_submodules(model, key):
    """
    Helper function to replace a module with transformers model
    https://github.com/huggingface/peft/blob/c0dd27bc974e4a62c6072142146887b75bb2de6c/src/peft/utils/other.py#L251
    """
    parent = model.get_submodule(".".join(key.split(".")[:-1]))
    target_name = key.split(".")[-1]
    target = model.get_submodule(key)
    return parent, target, target_name



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional

from logix import LogIX, LogIXScheduler
from logix.lightning.arguments import LogIXArguments
from logix.lightning.callback import LogIXCallback


def patch(
    module_cls,
    trainer_cls,
    logix_args: LogIXArguments,
    data_id_extractor: Optional[Callable] = None,
    mask_extractor: Optional[Callable] = None,
) -> LogIXCallback:
    run = LogIX(project=logix_args.project, config=logix_args.config)
    logix_scheduler = LogIXScheduler(
        run, hessian=logix_args.hessian, save=logix_args.save
    )

    logix_callback = LogIXCallback(
        logix=run,
        logix_scheduler=logix_scheduler,
        logix_args=logix_args,
        data_id_extractor=data_id_extractor,
        mask_extractor=mask_extractor,
    )

    class PatchedModule(module_cls):
        def __init__(self, *args, **kwargs) -> None:
            super().__init__(*args, **kwargs)

            self.logix = run
            self.logix_args = logix_args

            self.logix.watch(
                self._get_model(),
                name_filter=logix_args.name_filter,
                type_filter=logix_args.type_filter,
            )

            if self.logix_args.lora:
                self.logix.add_lora()

            if self.logix_args.initialize_from_log:
                run.initialize_from_log()

        def _get_model(self):
            return getattr(self, self.logix_args.model_key)

    class PatchedTrainer(trainer_cls):
        def __init__(
            self,
            accelerator="auto",
            strategy="auto",
            devices="auto",
            num_nodes=1,
            precision=None,
            logger=None,
            callbacks=None,
            fast_dev_run=False,
            max_epochs=None,
            min_epochs=None,
            max_steps=-1,
            min_steps=None,
            max_time=None,
            limit_train_batches=None,
            limit_val_batches=None,
            limit_test_batches=None,
            limit_predict_batches=None,
            overfit_batches=0.0,
            val_check_interval=None,
            check_val_every_n_epoch=1,
            num_sanity_val_steps=None,
            log_every_n_steps=None,
            enable_checkpointing=None,
            enable_progress_bar=None,
            enable_model_summary=None,
            accumulate_grad_batches=1,
            gradient_clip_val=None,
            gradient_clip_algorithm=None,
            deterministic=None,
            benchmark=None,
            inference_mode=True,
            use_distributed_sampler=True,
            profiler=None,
            detect_anomaly=False,
            barebones=False,
            plugins=None,
            sync_batchnorm=False,
            reload_dataloaders_every_n_epochs=0,
            default_root_dir=None,
        ) -> None:
            # add logix callback
            if callbacks is None:
                callbacks = [logix_callback]
            else:
                callbacks = callbacks + [logix_callback]

            # disable validation
            check_val_every_n_epoch = 1_000_000_000

            # disable checkpointing
            enable_checkpointing = False

            # disable logging
            logger = False

            super().__init__(
                accelerator,
                strategy,
                devices,
                num_nodes,
                precision,
                logger,
                callbacks,
                fast_dev_run,
                max_epochs,
                min_epochs,
                max_steps,
                min_steps,
                max_time,
                limit_train_batches,
                limit_val_batches,
                limit_test_batches,
                limit_predict_batches,
                overfit_batches,
                val_check_interval,
                check_val_every_n_epoch,
                num_sanity_val_steps,
                log_every_n_steps,
                enable_checkpointing,
                enable_progress_bar,
                enable_model_summary,
                accumulate_grad_batches,
                gradient_clip_val,
                gradient_clip_algorithm,
                deterministic,
                benchmark,
                inference_mode,
                use_distributed_sampler,
                profiler,
                detect_anomaly,
                barebones,
                plugins,
                sync_batchnorm,
                reload_dataloaders_every_n_epochs,
                default_root_dir,
            )

            self.logix = run
            self.logix_args = logix_args

        def extract_log(self, *args, **kwargs):
            self.logix_args.mode = "log"
            self.fit(*args, **kwargs)

        def influence(self, *args, **kwargs):
            self.logix_args.mode = "influence"
            self.fit(*args, **kwargs)

            return self.logix.influence.get_influence_scores()

        def self_influence(self, *args, **kwargs):
            self.logix_args.mode = "self_influence"
            self.fit(*args, **kwargs)

            return

    return PatchedModule, PatchedTrainer



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from contextlib import contextmanager
from functools import reduce

import numpy as np


def extract_arrays(obj, base_path=()):
    if isinstance(obj, dict):
        for k, v in obj.items():
            yield from extract_arrays(v, base_path + (k,))
    elif isinstance(obj, np.ndarray):
        yield base_path, obj


def get_from_nested_dict(nested_dict, keys):
    current_level = nested_dict
    for key in keys:
        if key in current_level:
            current_level = current_level[key]
    return current_level


class MemoryMapHandler:
    @staticmethod
    def write(save_path, filename, data_buffer, write_order_key, dtype="uint8"):
        file_root, file_ext = os.path.splitext(filename)
        if file_ext == "":
            filename += ".mmap"

        mmap_filename = os.path.join(save_path, filename)
        metadata_filename = os.path.join(save_path, file_root + "_metadata.json")

        total_size = sum(
            arr.nbytes for _, d in data_buffer for _, arr in extract_arrays(d)
        )
        mmap = np.memmap(mmap_filename, dtype=dtype, mode="w+", shape=(total_size,))

        metadata = []
        offset = 0
        for data_id, nested_dict in data_buffer:
            data_level_offset = offset
            data_shape = []
            block_size = 0
            # Enforcing the insert order based on the module path.
            for key in write_order_key:
                arr = get_from_nested_dict(nested_dict, key)
                bytes = arr.nbytes
                data_shape.append(arr.shape)
                block_size += reduce(lambda x, y: x * y, arr.shape)
                mmap[offset : offset + bytes] = arr.ravel().view(dtype)
                offset += arr.nbytes
            metadata.append(
                {
                    "data_id": data_id,
                    "path": write_order_key,
                    "offset": data_level_offset,
                    "shape": data_shape,
                    "block_size": block_size,
                    "dtype": str(arr.dtype),
                }
            )

        mmap.flush()
        del mmap  # Release the memmap object

        with open(metadata_filename, "w") as f:
            json.dump(metadata, f, indent=2)

    @staticmethod
    @contextmanager
    def read(path, filename, dtype="uint8"):
        """
        read reads the file by chunk index, it will return the data_buffer with metadata.
        Arg:
            filename (str): filename for the path to mmap.
        Returns:
            mmap (np.mmap): memory mapped buffer read from filename.
        """
        _, file_ext = os.path.splitext(filename)
        if file_ext == "":
            filename += ".mmap"

        mmap = np.memmap(os.path.join(path, filename), dtype=dtype, mode="r+")
        try:
            yield mmap
        finally:
            del mmap

    @staticmethod
    def read_metafile(path, meta_filename):
        _, file_ext = os.path.splitext(meta_filename)
        if file_ext == "":
            meta_filename += ".json"
        f = open(os.path.join(path, meta_filename), "r")
        metadata = json.load(f)  # This throws error when file does not exist.
        f.close()
        return metadata



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, List

from logix.statistic import CorrectedEigval, Covariance, Log, Mean, Variance
from logix.utils import get_logger

_PLUGIN_MAPPING = {
    "log": Log,
    "mean": Mean,
    "variance": Variance,
    "covariance": Covariance,
    "corrected_eigval": CorrectedEigval,
}
_PLUGIN_LIST = [Log, Mean, Variance, Covariance, CorrectedEigval]


def _reorder_plugins(plugins):
    """
    Reorder the plugins to ensure that the plugins are in the correct order. Especially,
    it is important to ensure that the Log plugin is the first plugin.
    Args:
        plugins: List of plugins.
    Returns:
        List of plugins in the correct order.
    """
    order = [Log, Mean, Variance, Covariance, CorrectedEigval]
    ordered_plugins = []
    for plugin in order:
        if plugin in plugins:
            ordered_plugins.append(plugin)
    return ordered_plugins


def _to_plugins(plugins: List[Any], is_grad: bool = False):
    """
    Convert and reorder the list of plugins to the actual plugins.
    """
    # Convert the string plugins to the actual plugins.
    for idx, plugin in enumerate(plugins):
        if isinstance(plugin, str):
            assert plugin in _PLUGIN_MAPPING
            plugins[idx] = _PLUGIN_MAPPING[plugin]
        assert plugins[idx] in _PLUGIN_LIST

    # reorder the plugins to ensure that the plugins are in the correct order.
    plugins = _reorder_plugins(plugins)

    if is_grad:
        # Ensure that the Log plugin is the first plugin.
        if len(plugins) > 0 and Log not in plugins:
            get_logger().warning(
                "The `Log` plugin is not in the list of plugins. "
                "The `Log` plugin will be inserted at the beginning of the list."
            )
            plugins.insert(0, Log)

    return plugins


class LogOption:
    def __init__(self):
        self.forward = []
        self.backward = []
        self.grad = []

        self.clear()

    def setup(self, log_option_kwargs):
        """
        Update logging configurations.

        Args:
            log: Logging configurations.
            save: Saving configurations.
            statistic: Statistic configurations.
        """
        self.clear()

        forward = log_option_kwargs.get("forward", [])
        backward = log_option_kwargs.get("backward", [])
        grad = log_option_kwargs.get("grad", [])

        self.forward = _to_plugins(forward)
        self.backward = _to_plugins(backward)
        self.grad = _to_plugins(grad)

    def clear(self):
        """
        Clear all logging configurations.
        """
        self.forward = []
        self.backward = []
        self.grad = []



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import logging as default_logging
import sys
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import torch
import torch.distributed as dist
import torch.nn as nn

_logger = None


class DistributedRankFilter(default_logging.Filter):
    """
    This is a logging filter which will filter out logs from all ranks
    in distributed training except for rank 0.
    """

    def filter(self, record):
        return get_rank() == 0


def get_logger() -> default_logging.Logger:
    """
    Get global logger.
    """
    global _logger
    if _logger:
        return _logger
    logger = default_logging.getLogger("AnaLog")
    log_format = default_logging.Formatter(
        "[%(asctime)s] [%(levelname)s] %(message)s", "%Y-%m-%d %H:%M:%S"
    )

    logger.propagate = False
    logger.setLevel(default_logging.INFO)
    ch = default_logging.StreamHandler(stream=sys.stdout)
    ch.setLevel(default_logging.INFO)
    ch.setFormatter(log_format)

    # Apply the rank filter to the handler
    rank_filter = DistributedRankFilter()
    ch.addFilter(rank_filter)

    logger.addHandler(ch)

    _logger = logger
    return _logger


def to_numpy(tensor) -> np.ndarray:
    """
    Convert a tensor to NumPy array.

    Args:
        tensor: The tensor to be converted.
    """
    if isinstance(tensor, np.ndarray):
        return tensor
    elif isinstance(tensor, torch.Tensor):
        return tensor.cpu().detach().numpy()
    else:
        raise ValueError("Unsupported tensor type. Supported libraries: NumPy, PyTorch")


def get_world_size() -> int:
    """
    Get the number of processes in the current distributed group.
    """
    if dist.is_initialized():
        return dist.get_world_size()
    else:
        return 0


def get_rank(group=None) -> int:
    """
    Get the rank of the current process in the current distributed group.

    Args:
        group (optional): The process group to work on. If not specified,
            the default process group will be used.
    """
    if dist.is_initialized():
        return dist.get_rank(group)
    else:
        return 0


def get_repr_dim(named_modules) -> Tuple[List[str], List[int]]:
    repr_dims = []
    paths = []
    for k, v in named_modules.items():
        get_logger().info(f"{v}: {k}")
        repr_dims.append(k.weight.data.numel())
        paths.append((v, "grad"))  # hardcoded
    return paths, repr_dims


def print_tracked_modules(repr_dim) -> None:
    """
    Print the tracked modules.
    """
    get_logger().info("Tracking the following modules:")
    get_logger().info(f"Total number of parameters: {repr_dim:,}\n")


def module_check(
    module: nn.Module,
    module_name: str,
    supported_modules: Optional[List[nn.Module]] = None,
    type_filter: Optional[List[nn.Module]] = None,
    name_filter: Optional[List[str]] = None,
    is_lora: bool = False,
) -> bool:
    """
    Check if the module is supported for logging.

    Args:
        module (nn.Module): The module to check.
        module_name (str): Name of the module.
        supported_modules (Optional[List[nn.Module]]): List of supported module types.
        type_filter (Optional[List[nn.Module]]): List of module types to filter.
        name_filter (Optional[List[str]]): List of keywords to filter module names.
        is_lora (bool): Flag to check for specific 'analog_lora_B' in module names.

    Returns:
        bool: True if module is supported, False otherwise.
    """
    if list(module.children()):
        return False
    if supported_modules and not isinstance(module, tuple(supported_modules)):
        return False
    if type_filter and not isinstance(module, tuple(type_filter)):
        return False
    if name_filter and not any(keyword in module_name for keyword in name_filter):
        return False
    if is_lora and "logix_lora_B" not in module_name:
        return False
    return True


def nested_dict():
    """
    Helper function to create a nested defaultdict.
    """
    return defaultdict(nested_dict)


def merge_log_dict(merged_log_dict, log_dict) -> None:
    for key, value in log_dict.items():
        if isinstance(value, dict):
            merge_log_dict(merged_log_dict[key], value)
        else:
            if key not in merged_log_dict:
                merged_log_dict[key] = value
            else:
                merged_log_dict[key] = torch.cat([merged_log_dict[key], value], dim=0)


def merge_logs(log_list):
    merged_data_id = []
    merged_log_dict = nested_dict()

    for data_id, log_dict in log_list:
        merged_data_id.extend(data_id)
        merge_log_dict(merged_log_dict, log_dict)
    return merged_data_id, merged_log_dict


def flatten_log(log, path) -> torch.Tensor:
    flat_log_list = []
    for module, log_type in path:
        log_module = log[module][log_type]
        bsz = log_module.shape[0]
        flat_log_list.append(log_module.reshape(bsz, -1))
    flat_log = torch.cat(flat_log_list, dim=1)

    return flat_log


def unflatten_log(log, path):
    raise NotImplementedError


def synchronize_device(
    src: Dict[str, Dict[str, torch.Tensor]],
    tgt: Dict[str, Dict[str, torch.Tensor]],
    device: Optional[torch.device] = None,
) -> None:
    """
    Synchronize the device of two tensor dicts.

    Args:
        src (Dict[str, Dict[str, torch.Tensor]]): Source tensors
        tgt (Dict[str, Dict[str, torch.Tensor]]): Target tensors
        device (Optional[torch.device]): Device to synchronize to
    """

    for module_name, module_dict in tgt.items():
        for log in module_dict.keys():
            if device is None:
                device = src[module_name][log].device
            tgt[module_name][log] = tgt[module_name][log].to(device=device)


class DataIDGenerator:
    """
    Generate unique IDs for data.
    """

    def __init__(self, mode="hash") -> None:
        self.mode = mode
        if mode == "index":
            self.count = 0

    def __call__(self, data: Any):
        if self.mode == "hash":
            return self.generate_hash_id(data)
        elif self.mode == "index":
            return self.generate_index_id(data)
        else:
            raise ValueError(f"Unsupported mode: {self.mode}")

    def generate_hash_id(self, data: Any) -> List[str]:
        """
        Given data, generate id using SHA256 hash.
        """
        data_id = []
        for d in data:
            ndarray = to_numpy(d)
            ndarray.flags.writeable = False
            data_id.append(hashlib.sha256(ndarray.tobytes()).hexdigest())
        return data_id

    def generate_index_id(self, data: Any) -> List[str]:
        """
        Given data, generate id based on the index.
        """
        data_id = np.arange(self.count, self.count + len(data))
        data_id = [str(d) for d in data_id]
        self.count += len(data)
        return data_id



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Any, Dict, Optional, Tuple

import torch
import torch.nn as nn

from logix.batch_info import BatchInfo
from logix.config import LoggingConfig
from logix.logging.log_saver import LogSaver
from logix.logging.option import LogOption
from logix.logging.utils import compute_per_sample_gradient
from logix.state import LogIXState
from logix.statistic import Log


class HookLogger:
    def __init__(
        self,
        config: LoggingConfig,
        state: LogIXState,
        binfo: BatchInfo,
    ) -> None:
        """
        Initializes the LoggingHandler with empty lists for hooks.
        """
        self.state = state
        self.binfo = binfo
        self.opt = LogOption()

        # parse config
        self.cpu_offload = config.cpu_offload
        self.dtype = config.get_dtype()

        # log saver
        self.log_saver = LogSaver(config=config, state=self.state)

        # hooks
        self.modules_to_hook = []
        self.modules_to_name = {}
        self.forward_hooks = []
        self.backward_hooks = []
        self.grad_hooks = []
        self.tensor_hooks = []

    def log(self, data_id: Any, mask: Optional[torch.Tensor] = None):
        """
        Add log state on exit.

        Args:
            data_id: The data ID associated with the current batch.
            mask (torch.Tensor): A mask to be applied to the activations.
        """
        log = self.binfo.log.copy()
        self.binfo.clear()
        self.binfo.data_id = data_id
        self.binfo.mask = mask

        self.update()

        return log

    def save_log(self):
        # save log to disk
        self.log_saver.buffer_write(binfo=self.binfo)
        self.log_saver.flush()

    def update(self, save: bool = False):
        # gradient plugin has to be excecuted after accumulating all gradients
        for stat in self.opt.grad[1:]:
            for module_name, _ in self.binfo.log.items():
                stat.update(
                    state=self.state,
                    binfo=self.binfo,
                    module=None,
                    module_name=module_name,
                    log_type="grad",
                    data=None,
                    cpu_offload=self.cpu_offload,
                )

        # Wait for all asynchronous CUDA operations to finish
        if torch.cuda.is_available():
            torch.cuda.current_stream().synchronize()

        # Write and flush the buffer if necessary
        if save:
            self.save_log()

    def _forward_hook_fn(
        self, module: nn.Module, inputs: Tuple[torch.Tensor], module_name: str
    ) -> None:
        """
        Internal forward hook function.

        Args:
            module: The module triggering the hook.
            inputs: The input to the module.
            module_name (str): The name of the module.
        """
        assert len(inputs) == 1

        activations = inputs[0]

        # If `mask` is not None, apply the mask to activations. This is
        # useful for example when you work with sequence models that use
        # padding. In this case, you can use the mask to ignore the padding
        if self.binfo.mask is not None:
            mask = self.binfo.mask
            if len(mask.shape) != len(activations.shape):
                assert len(mask.shape) == len(activations.shape) - 1
                if mask.shape[-1] == activations.shape[-2]:
                    activations = activations * mask.unsqueeze(-1)
            else:
                if mask.shape[-1] == activations.shape[-1]:
                    activations = activations * mask

        if self.dtype is not None:
            activations = activations.to(dtype=self.dtype)

        for plugin in self.opt.forward:
            plugin.update(
                state=self.state,
                binfo=self.binfo,
                module=module,
                module_name=module_name,
                log_type="forward",
                data=activations,
                cpu_offload=self.cpu_offload,
            )

    def _backward_hook_fn(
        self,
        module: nn.Module,
        grad_inputs: Tuple[torch.Tensor],
        grad_outputs: Tuple[torch.Tensor],
        module_name: str,
    ) -> None:
        """
        Internal backward hook function.

        Args:
            module: The module triggering the hook.
            grad_inputs: The gradient of the input to the module.
            grad_outputs: The gradient of the output from the module.
            module_name (str): The name of the module.
        """
        assert len(grad_outputs) == 1

        error = grad_outputs[0]

        if self.dtype is not None:
            error = error.to(dtype=self.dtype)

        for plugin in self.opt.backward:
            plugin.update(
                state=self.state,
                binfo=self.binfo,
                module=module,
                module_name=module_name,
                log_type="backward",
                data=error,
                cpu_offload=self.cpu_offload,
            )

    def _grad_hook_fn(
        self,
        module: nn.Module,
        inputs: Tuple[torch.Tensor],
        outputs: Tuple[torch.Tensor],
        module_name: str,
    ) -> None:
        """
        Internal gradient hook function.

        Args:
            module: The module triggering the hook.
            inputs: The input to the module.
            outputs: The output from the module.
            module_name (str): The name of the module.
        """
        assert len(inputs) == 1

        # In case, the same module is used multiple times in the forward pass,
        # we need to accumulate the gradients. We achieve this by using the
        # additional tensor hook on the output of the module.
        def _grad_backward_hook_fn(grad: torch.Tensor):
            if len(self.opt.grad) > 0:
                assert self.opt.grad[0] == Log
                per_sample_gradient = compute_per_sample_gradient(
                    inputs[0], grad, module
                )

                if self.dtype is not None:
                    per_sample_gradient = per_sample_gradient.to(dtype=self.dtype)

                for plugin in self.opt.grad[:1]:
                    plugin.update(
                        state=self.state,
                        binfo=self.binfo,
                        module=module,
                        module_name=module_name,
                        log_type="grad",
                        data=per_sample_gradient,
                        cpu_offload=self.cpu_offload,
                    )

        tensor_hook = outputs.register_hook(_grad_backward_hook_fn)
        self.tensor_hooks.append(tensor_hook)

    def _tensor_forward_hook_fn(self, tensor: torch.Tensor, tensor_name: str) -> None:
        """
        Internal forward hook function specifically designed for tensors.

        This method allows you to track the activations of tensors that are not
        necessarily tied to any specific module, but are still of interest.

        Args:
            tensor: The tensor triggering the hook.
            tensor_name (str): A string identifier for the tensor, useful for logging.
        """
        if self.dtype is not None:
            tensor = tensor.to(dtype=self.dtype)

        for plugin in self.opt.forward:
            plugin.update(
                state=self.state,
                binfo=self.binfo,
                module=None,
                module_name=tensor_name,
                log_type="forward",
                data=tensor,
                cpu_offload=self.cpu_offload,
            )

    def _tensor_backward_hook_fn(self, grad: torch.Tensor, tensor_name: str) -> None:
        """
        Internal backward hook function specifically designed for tensors.

        This method allows you to track the gradients associated with specific
        tensors that are not necessarily tied to any specific module, but are still of interest.

        Args:
            grad: The gradient tensor triggering the hook.
            tensor_name (str): A string identifier for the tensor whose gradient is being tracked.
        """
        if self.dtype is not None:
            grad = grad.to(dtype=self.dtype)

        for plugin in self.opt.backward:
            plugin.update(
                state=self.state,
                binfo=self.binfo,
                module=None,
                module_name=tensor_name,
                log_type="backward",
                data=grad,
                cpu_offload=self.cpu_offload,
            )

    def register_all_module_hooks(self) -> None:
        """
        Register all module hooks.
        """
        for module in self.modules_to_hook:
            # As each hook has its own function, we need to register all hooks
            # separately. We use partial functions to pass the module name to
            # the hook functions.
            module_name = self.modules_to_name[module]
            forward_hook = module.register_forward_pre_hook(
                partial(self._forward_hook_fn, module_name=module_name)
            )
            backward_hook = module.register_full_backward_hook(
                partial(self._backward_hook_fn, module_name=module_name)
            )
            grad_hook = module.register_forward_hook(
                partial(self._grad_hook_fn, module_name=module_name)
            )
            self.forward_hooks.append(forward_hook)
            self.backward_hooks.append(backward_hook)
            self.grad_hooks.append(grad_hook)

    def register_all_tensor_hooks(self, tensor_dict: Dict[str, torch.Tensor]) -> None:
        """
        Register all tensor hooks.

        Args:
            tensor_dict (dict): Dictionary containing tensor names as keys and tensors as values.
        """
        for tensor_name, tensor in tensor_dict.items():
            self._tensor_forward_hook_fn(tensor, tensor_name)
            tensor_hook = tensor.register_hook(
                partial(self._tensor_backward_hook_fn, tensor_name=tensor_name)
            )
            self.tensor_hooks.append(tensor_hook)

    def add_module(self, module_name: str, module: nn.Module) -> None:
        """
        Add a module to be hooked.

        Args:
            module_name (str): The name of the module.
            module: The module to be hooked.
        """
        self.modules_to_hook.append(module)
        self.modules_to_name[module] = module_name

    def setup(self, log_option_kwargs: Dict[str, Any]) -> None:
        """
        Update logging configurations.

        Args:
            log_option_kwargs: Logging configurations.
        """
        self.opt.setup(log_option_kwargs)

    def finalize(self):
        """
        Dump everything in the buffer to a disk.
        """
        self.log_saver.finalize()

    def clear(
        self, hook: bool = True, module: bool = True, buffer: bool = True
    ) -> None:
        """
        Clear all hooks and internal states.

        Args:
            hook (bool): Whether to clear the hooks.
            module (bool): Whether to clear the modules.
        """
        if hook:
            for hook in self.forward_hooks:
                hook.remove()
            self.forward_hooks.clear()
            for hook in self.backward_hooks:
                hook.remove()
            self.backward_hooks.clear()
            for hook in self.grad_hooks:
                hook.remove()
            self.grad_hooks.clear()
            for hook in self.tensor_hooks:
                hook.remove()
            self.tensor_hooks.clear()

        if module:
            self.modules_to_hook = []
            self.modules_to_name = {}

        if buffer:
            self.log_saver.buffer_clear()



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
from typing import Dict, Optional, Tuple

import torch
import torch.distributed as dist

from logix.utils import get_rank, get_world_size, nested_dict


class LogIXState:
    """
    LogIXState stores all these relevant log states that are used for
    communication between different handlers. All states in LogIXState
    follow the nested dictionary data structure.
    """

    def __init__(self) -> None:
        self._states = set()
        self._states_to_synchronize = []
        self._states_to_save = []
        self._states_to_normalize = []
        self._states_to_not_clear = set()

        self.register_state("mean_state", synchronize=True, save=True)
        self.register_state("mean_counter", synchronize=True, save=False)
        self.register_state("covariance_state", synchronize=True, save=True)
        self.register_state("covariance_counter", synchronize=True, save=False)
        self.register_state(
            "model_module", synchronize=False, save=True, not_clear=True
        )

        self.register_normalize_pair("mean_state", "mean_counter")
        self.register_normalize_pair("covariance_state", "covariance_counter")

    def register_state(
        self,
        state_name: str,
        synchronize: bool = False,
        save: bool = False,
        not_clear: bool = False,
    ):
        """
        Register a state to be logged.

        Args:
            state_name: Name of the state.
            synchronize: Whether to synchronize the state across processes.
            save: Whether to save the state to disk.
            not_clear: Whether to clear the state.
        """
        if state_name in self._states:
            return

        assert state_name not in self._states
        setattr(self, state_name, nested_dict())
        self._states.add(state_name)
        if synchronize:
            self._states_to_synchronize.append(state_name)
        if save:
            self._states_to_save.append(state_name)
        if not_clear:
            self._states_to_not_clear.add(state_name)

    def register_normalize_pair(self, state_name: str, counter_name: str):
        """
        Add a normalization pair to the state.

        Args:
            state_name: Name of the state.
            counter_name: Name of the counter.
        """
        if (state_name, counter_name) in self._states_to_normalize:
            return

        self._states_to_normalize.append((state_name, counter_name))

    @torch.no_grad()
    def covariance_svd(self) -> None:
        """
        Compute the SVD of the covariance.
        """
        self.register_state("covariance_eigval_state", save=True)
        self.register_state("covariance_eigvec_state", save=True)

        for module_name, module_state in self.covariance_state.items():
            for mode, covariance in module_state.items():
                dtype = covariance.dtype
                eigvals, eigvecs = torch.linalg.eigh(covariance.double())
                self.covariance_eigval_state[module_name][mode] = eigvals.to(
                    dtype=dtype
                )
                self.covariance_eigvec_state[module_name][mode] = eigvecs.to(
                    dtype=dtype
                )

    @torch.no_grad()
    def covariance_inverse(self, damping: Optional[float] = None) -> None:
        """
        Compute the inverse of the covariance.

        Args:
            damping: Damping for the covariance.
        """
        self.register_state("covariance_inverse_state", save=True)

        for module_name, module_state in self.covariance_state.items():
            for mode, cov in module_state.items():
                damping_module = (
                    0.1 * torch.trace(cov) / cov.size(0) if damping is None else damping
                )
                self.covariance_inverse_state[module_name][mode] = torch.inverse(
                    cov + damping_module * torch.eye(cov.size(0)).to(device=cov.device)
                )

    def get_covariance_state(self) -> Dict[str, Dict[str, torch.Tensor]]:
        """
        Return the covariance state.
        """
        return self.covariance_state

    def get_covariance_inverse_state(
        self, damping: Optional[float] = None
    ) -> Dict[str, Dict[str, torch.Tensor]]:
        """
        Return the covariance inverse state. If the state is not computed, compute
        it first.
        """
        if not hasattr(self, "covariance_inverse_state"):
            self.covariance_inverse(damping=damping)
        return self.covariance_inverse_state

    def get_covariance_svd_state(
        self,
    ) -> Tuple[Dict[str, Dict[str, torch.Tensor]], Dict[str, Dict[str, torch.Tensor]]]:
        """
        Return the covariance SVD state. If the state is not computed, compute
        it first.
        """
        if not hasattr(self, "covariance_eigval_state") or not hasattr(
            self, "covariance_eigvec_state"
        ):
            self.covariance_svd()
        if hasattr(self, "ekfac_eigval_state"):
            return self.ekfac_eigval_state, self.covariance_eigvec_state
        return self.covariance_eigval_state, self.covariance_eigvec_state

    def synchronize(self) -> None:
        """
        Synchronize all synchronizable states across processes.
        """

        # synchronize helper function
        def _synchronize(state_dict):
            for key in state_dict:
                if isinstance(state_dict[key], dict):
                    _synchronize(state_dict[key])
                else:
                    # we need to move the state to the GPU before all-reducing
                    # across processes as the communication is only
                    # supported on GPU tensors in most PyTorch backends.
                    if not isinstance(state_dict[key], torch.Tensor):
                        state_dict[key] = torch.tensor(state_dict[key])
                    state_gpu = state_dict[key].cuda()
                    dist.all_reduce(state_gpu, op=dist.ReduceOp.SUM)
                    state_dict[key].copy_(state_gpu.cpu())

                    torch.cuda.synchronize()

        for state_name in self._states_to_synchronize:
            state_dict = getattr(self, state_name)
            _synchronize(state_dict)

    def normalize(self) -> None:
        """
        Normalize the state.
        """

        # normalize helper function
        def _normalize(state_dict, counter_dict):
            for key in state_dict:
                if not isinstance(state_dict[key], torch.Tensor):
                    _normalize(state_dict[key], counter_dict[key])
                else:
                    state_dict[key] /= counter_dict[key]

        for state_name, counter_name in self._states_to_normalize:
            state_dict = getattr(self, state_name)
            counter_dict = getattr(self, counter_name)
            _normalize(state_dict, counter_dict)

    def finalize(self, log_dir: str) -> None:
        """
        Finalize the state.

        Args:
            log_dir: Directory to save the state.
        """
        # If we are in distributed mode, synchronize the states across processes
        if get_world_size() > 1:
            self.synchronize()

        # Normalize the states with the counter
        self.normalize()

        # Save the states to disk
        if get_rank() == 0:
            self.save_state(log_dir=log_dir)

    def save_state(self, log_dir: str) -> None:
        """
        Save state to disk.
        """
        state_log_dir = os.path.join(log_dir, "state")
        if not os.path.exists(state_log_dir) and get_rank() == 0:
            os.makedirs(state_log_dir)

        for state_name in self._states_to_save:
            state_dict = getattr(self, state_name)
            torch.save(state_dict, os.path.join(state_log_dir, f"{state_name}.pt"))

        others = {
            "_states_to_synchronize": self._states_to_synchronize,
            "_states_to_save": self._states_to_save,
            "_states_to_normalize": self._states_to_normalize,
        }
        torch.save(others, os.path.join(state_log_dir, "others.pt"))

    def load_state(self, log_dir: str) -> None:
        """
        Load state from disk.
        """
        state_log_dir = os.path.join(log_dir, "state")

        others = torch.load(os.path.join(state_log_dir, "others.pt"))
        for key, value in others.items():
            setattr(self, key, value)

        for state_name in self._states_to_save:
            state_dict = torch.load(os.path.join(state_log_dir, f"{state_name}.pt"))
            setattr(self, state_name, state_dict)

    def set_state(self, state_name: str, **kwargs) -> None:
        """
        set_state sets the state for the given state_name with input kwargs.
        """
        if state_name not in self._states:
            raise ValueError("state name {} not in registered state".format(state_name))
        for key, value in kwargs.items():
            state = getattr(self, state_name)
            state[key] = value

    def get_state(self, state_name: str) -> Dict[str, Dict[str, torch.Tensor]]:
        return getattr(self, state_name)

    def clear(self) -> None:
        """
        Clear all states.
        """
        for state_name in self._states:
            if state_name in self._states_to_not_clear:
                continue
            state = getattr(self, state_name)
            state.clear()



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import torch
import torch.nn as nn

from logix.lora.utils import compute_top_k_singular_vectors


class LoraLinear(nn.Module):
    def __init__(self, rank: int, linear: nn.Linear, shared_module: nn.Linear = None):
        """Transforms a linear layer into a LoraLinear layer.

        Args:
            rank (int): The rank of lora
            linear (nn.Linear): The linear layer to transform
        """
        super().__init__()

        in_features = linear.in_features
        out_features = linear.out_features

        self.rank = min(rank, in_features, out_features)

        self.logix_lora_A = nn.Linear(in_features, self.rank, bias=False)
        self.logix_lora_B = shared_module or nn.Linear(self.rank, self.rank, bias=False)
        self.logix_lora_C = nn.Linear(self.rank, out_features, bias=False)

        nn.init.kaiming_uniform_(self.logix_lora_A.weight, a=math.sqrt(5))
        nn.init.zeros_(self.logix_lora_B.weight)
        nn.init.kaiming_uniform_(self.logix_lora_C.weight, a=math.sqrt(5))

        self._linear = linear

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        result = self._linear(input)
        result += self.logix_lora_C(self.logix_lora_B(self.logix_lora_A(input)))

        return result

    def pca_init_weight(self, covariance=None):
        """Initialize the weight of the LoraLinear layer.

        Args:
            init_strategy (str): The type of projection to use
            covariance (dict): The forward and backward covariance of the layer
        """
        (
            top_r_singular_vector_forward,
            top_r_singular_value_forward,
        ) = compute_top_k_singular_vectors(covariance["forward"], self.rank)
        (
            top_r_singular_vector_backward,
            top_r_singular_value_backward,
        ) = compute_top_k_singular_vectors(covariance["backward"], self.rank)
        self.logix_lora_A.weight.data.copy_(top_r_singular_vector_forward.T)
        self.logix_lora_C.weight.data.copy_(top_r_singular_vector_backward)


class LoraConv2d(nn.Module):
    def __init__(self, rank: int, conv: nn.Conv2d, shared_module: nn.Conv2d = None):
        """Transforms a conv2d layer into a LoraConv2d layer.

        Args:
            rank (int): The rank of lora
            conv (nn.Conv2d): The conv2d layer to transform
        """
        super().__init__()

        in_channels = conv.in_channels
        out_channels = conv.out_channels
        kernel_size = conv.kernel_size
        stride = conv.stride
        padding = conv.padding

        self.rank = min(rank, in_channels, out_channels)

        self.logix_lora_A = nn.Conv2d(
            in_channels, self.rank, kernel_size, stride, padding, bias=False
        )
        self.logix_lora_B = shared_module or nn.Conv2d(
            self.rank, self.rank, 1, bias=False
        )
        self.logix_lora_C = nn.Conv2d(self.rank, out_channels, 1, bias=False)

        nn.init.kaiming_uniform_(self.logix_lora_A.weight, a=math.sqrt(5))
        nn.init.zeros_(self.logix_lora_B.weight)
        nn.init.kaiming_uniform_(self.logix_lora_C.weight, a=math.sqrt(5))

        self._conv = conv

    def forward(self, input) -> torch.Tensor:
        result = self._conv(input)
        result += self.logix_lora_C(self.logix_lora_B(self.logix_lora_A(input)))

        return result

    def pca_init_weight(self, covariance):
        """Initialize the weight of the LoraConv2d layer.

        Args:
            projection_type (str): The type of projection to use
            covariance (dict): The forward and backward covariance of the layer
        """
        (
            top_r_singular_vector_forward,
            top_r_singular_value_forward,
        ) = compute_top_k_singular_vectors(covariance["forward"], self.rank)
        (
            top_r_singular_vector_backward,
            top_r_singular_value_backward,
        ) = compute_top_k_singular_vectors(covariance["backward"], self.rank)
        shape_A = self.logix_lora_A.weight.shape
        shape_C = self.logix_lora_C.weight.shape
        self.logix_lora_A.weight.data.copy_(
            top_r_singular_vector_forward.T.view(shape_A)
        )
        self.logix_lora_C.weight.data.copy_(
            top_r_singular_vector_backward.view(shape_C)
        )


class LoraEmbedding(nn.Module):
    def __init__(
        self, rank: int, embedding: nn.Embedding, shared_module: nn.Embedding = None
    ):
        """Transforms a linear layer into a LoraEmbedding layer.

        Args:
            rank (int): The rank of lora
            linear (nn.Linear): The linear layer to transform
        """
        super().__init__()

        num_embeddings = embedding.num_embeddings
        embedding_dim = embedding.embedding_dim

        self.rank = min(rank, num_embeddings, embedding_dim)

        self.logix_lora_A = nn.Embedding(num_embeddings, self.rank)
        self.logix_lora_B = shared_module or nn.Linear(self.rank, self.rank, bias=False)
        self.logix_lora_C = nn.Linear(self.rank, embedding_dim, bias=False)

        nn.init.kaiming_uniform_(self.logix_lora_A.weight, a=math.sqrt(5))
        nn.init.zeros_(self.logix_lora_B.weight)
        nn.init.kaiming_uniform_(self.logix_lora_C.weight, a=math.sqrt(5))

        self._embedding = embedding

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        result = self._embedding(input)
        result += self.logix_lora_C(self.logix_lora_B(self.logix_lora_A(input)))

        return result

    def pca_init_weight(self, covariance=None):
        """Initialize the weight of the LoraEmbedding layer.

        Args:
            init_strategy (str): The type of projection to use
            covariance (dict): The forward and backward covariance of the layer
        """
        raise NotImplementedError



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F


class InvalidModuleError(Exception):
    """
    Raised when the provided module is invalid.
    """

    pass


def extract_patches(
    inputs: torch.Tensor,
    kernel_size: Tuple[int, int],
    stride: Tuple[int, int],
    padding: Tuple[int, int],
) -> torch.Tensor:
    """Extract patches for the KFC approximation.

    This method is based on the technique described in https://arxiv.org/pdf/1602.01407.pdf.

    Args:
        inputs (torch.Tensor):
            Activations before the convolutional layer.
        kernel_size (tuple):
            Dimensions of the convolutional layer's kernel.
        stride (tuple):
            Stride applied in the convolutional layer.
        padding (tuple):
            Padding dimensions applied in the convolutional layer.
    """
    if padding[0] + padding[1] > 0:
        inputs = F.pad(inputs, (padding[1], padding[1], padding[0], padding[0])).data
    inputs = inputs.unfold(2, kernel_size[0], stride[0])
    inputs = inputs.unfold(3, kernel_size[1], stride[1])
    inputs = inputs.transpose_(1, 2).transpose_(2, 3).contiguous()
    inputs = inputs.view(
        inputs.size(0),
        inputs.size(1),
        inputs.size(2),
        inputs.size(3) * inputs.size(4) * inputs.size(5),
    )
    return inputs


def make_forward_2d(
    data: torch.Tensor,
    module: nn.Module,
) -> torch.Tensor:
    """Extract and reshape activations into valid shapes for covariance computations.

    Args:
        activations (torch.Tensor):
            Raw pre-activations supplied to the module.
        module (nn.Module):
            The module where the activations are applied.
    """
    if isinstance(module, nn.Linear):
        reshaped_data = data.reshape(-1, data.shape[-1])
    elif isinstance(module, nn.Conv2d):
        reshaped_data = extract_patches(
            data, module.kernel_size, module.stride, module.padding
        )
        reshaped_data = reshaped_data.view(-1, reshaped_data.size(-1))
    else:
        raise InvalidModuleError()
    return reshaped_data


def make_backward_2d(
    data: torch.Tensor,
    module: nn.Module,
) -> torch.Tensor:
    """Extract and reshape gradients into valid shapes for covariance computations.

    Args:
        gradients (torch.Tensor):
            Raw gradients on the output to the module.
        module (nn.Module):
            The module where the gradients are computed.
    """
    if isinstance(module, nn.Linear):
        del module
        reshaped_data = data.reshape(-1, data.shape[-1])
    elif isinstance(module, nn.Conv2d):
        del module
        reshaped_data = data.permute(0, 2, 3, 1)
        reshaped_data = reshaped_data.reshape(-1, reshaped_data.size(-1))
    else:
        raise InvalidModuleError()
    return reshaped_data


def make_2d(data: torch.Tensor, module: nn.Module, log_type: str) -> torch.Tensor:
    """Extract and reshape data into 2d for computing statistic.

    Args:
        module (nn.Module):
            The module where the activations are applied.
        mode (str):
            Forward or backward.
        data (torch.Tensor):
            Activations corresponding to the module.
    """
    if module is None:
        return data.reshape(data.shape[0], -1)
    elif log_type == "forward":
        return make_forward_2d(data, module)
    elif log_type == "backward":
        return make_backward_2d(data, module)
    else:
        raise ValueError(f"Invalid mode {log_type}")



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
import torch.nn as nn

from logix.batch_info import BatchInfo
from logix.state import LogIXState


class CorrectedEigval:
    @staticmethod
    @torch.no_grad()
    def update(
        state: LogIXState,
        binfo: BatchInfo,
        module: nn.Module,
        module_name: str,
        log_type: str,
        data: Optional[torch.Tensor] = None,
        cpu_offload: Optional[bool] = False,
    ):
        """
        Update the mean state.
        """
        if not hasattr(state, "covariance_eigvec_state"):
            state.covariance_svd()
            state.register_state("ekfac_eigval_state", synchronize=True, save=True)
            state.register_state("ekfac_counter", synchronize=True, save=False)
            state.register_normalize_pair("ekfac_eigval_state", "ekfac_counter")
        covariance_eigvec_state = state.covariance_eigvec_state
        covariance_eigval_state = state.covariance_eigval_state
        ekfac_eigval_state = state.ekfac_eigval_state
        ekfac_counter = state.ekfac_counter

        assert data is None
        data = binfo.log[module_name]["grad"]

        if module_name not in ekfac_eigval_state:
            device = data.device if not cpu_offload else "cpu"
            dtype = data.dtype
            ekfac_eigval_state[module_name] = torch.zeros(
                data.shape[-2], data.shape[-1], device=device, dtype=dtype
            )
            ekfac_counter[module_name] = 0

        data = data.detach()
        if cpu_offload:
            eigvec_fwd_gpu = covariance_eigvec_state[module_name]["forward"].to(
                device=data.device
            )
            eigvec_bwd_gpu = covariance_eigvec_state[module_name]["backward"].to(
                device=data.device
            )
            ekfac_eigval_state_gpu = ekfac_eigval_state[module_name].to(
                device=data.device
            )
            rotated_grads = torch.matmul(data, eigvec_fwd_gpu)
            for rotated_grad in rotated_grads:
                weight = torch.matmul(eigvec_bwd_gpu.t(), rotated_grad)
                ekfac_eigval_state_gpu.add_(weight.square_())
            ekfac_eigval_state[module_name] = ekfac_eigval_state_gpu.to(
                device="cpu", non_blocking=True
            )

            # TODO: Not sure if this improves memory usage.
            del eigvec_fwd_gpu
            del eigvec_bwd_gpu
            del ekfac_eigval_state_gpu
        else:
            rotated_grads = torch.matmul(
                data, covariance_eigvec_state[module_name]["forward"]
            )
            for rotated_grad in rotated_grads:
                weight = torch.matmul(
                    covariance_eigvec_state[module_name]["backward"].t(), rotated_grad
                )
                ekfac_eigval_state[module_name].add_(weight.square_())

        ekfac_counter[module_name] += len(data)



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.nn as nn
from einops import einsum, reduce


@torch.no_grad()
def compute_per_sample_gradient(
    fwd: torch.Tensor, bwd: torch.Tensor, module: nn.Module
):
    """
    Computes the per-sample gradient of a module.

    Args:
        fwd: The forward activations of the module.
        bwd: The backward activations of the module.
        module: The module whose per-sample gradient needs to be computed.
    """
    if isinstance(module, nn.Linear):
        # For linear layers, we can simply compute the outer product of the
        # forward and backward activations.
        outer_product = einsum(bwd, fwd, "... i, ... j -> ... i j")
        return reduce(outer_product, "n ... i j -> n i j", "sum")
    elif isinstance(module, nn.Conv2d):
        # For convolutional layers, we need to unfold the forward activations
        # and compute the outer product of the backward and unfolded forward
        # activations.
        bsz = fwd.shape[0]
        fwd_unfold = torch.nn.functional.unfold(
            fwd,
            module.kernel_size,
            dilation=module.dilation,
            padding=module.padding,
            stride=module.stride,
        )
        fwd_unfold = fwd_unfold.reshape(bsz, fwd_unfold.shape[1], -1)
        bwd = bwd.reshape(bsz, -1, fwd_unfold.shape[-1])
        grad = einsum(bwd, fwd_unfold, "i j k, i l k -> i j l")

        # Ensure that each gradient has two dimensions of (out_dim, in_dim)
        shape = [bsz, module.weight.shape[0], -1]
        return grad.reshape(shape)
    elif isinstance(module, nn.Conv1d):
        raise NotImplementedError
    else:
        raise ValueError(f"Unsupported module type: {type(module)}")



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
import torch.nn as nn

from logix.batch_info import BatchInfo
from logix.state import LogIXState


class Log:
    @staticmethod
    @torch.no_grad()
    def update(
        state: LogIXState,
        binfo: BatchInfo,
        module: nn.Module,
        module_name: str,
        log_type: str,
        data: Optional[torch.Tensor] = None,
        cpu_offload: Optional[bool] = False,
    ):
        """
        Put log into `binfo`
        """
        module_log = binfo.log[module_name]
        if log_type not in module_log:
            module_log[log_type] = data
        else:
            module_log[log_type] += data



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import torch
import torch.nn as nn

from logix.batch_info import BatchInfo
from logix.state import LogIXState
from logix.statistic.utils import make_2d


class Variance:
    @staticmethod
    @torch.no_grad()
    def update(
        state: LogIXState,
        binfo: BatchInfo,
        module: nn.Module,
        module_name: str,
        log_type: str,
        data: Optional[torch.Tensor] = None,
        cpu_offload: Optional[bool] = False,
    ):
        """
        Update the variance state.
        """
        variance_state = state.variance_state
        variance_counter = state.variance_counter
        if data is None:
            data = binfo.log[module_name][log_type]

        # extract and reshape data to 2d tensor for variance computation
        batch_size = data.size(0)
        data = make_2d(data, module, log_type).detach()

        # initialize variance state if necessary
        if log_type not in variance_state[module_name]:
            device = data.device if not cpu_offload else "cpu"
            dtype = data.dtype
            variance_state[module_name][log_type] = torch.zeros(
                data.shape[-1], device=device, dtype=dtype
            )
            variance_counter[module_name][log_type] = 0

        # update variance state
        if cpu_offload:
            # By default, all states are stored on the CPU, and therefore
            # computing updates for states on CPU is slow. For efficiency,
            # we move states to the GPU if data is on the GPU, and then
            # move it back to the CPU asynchrously.
            variance_state_gpu = variance_state[module_name][log_type].to(
                device=data.device
            )
            variance_state_gpu.add_(data.sum(dim=0))
            variance_state[module_name][log_type] = variance_state_gpu.to(
                device="cpu", non_blocking=True
            )
        else:
            variance_state[module_name][log_type].add_(data.sum(dim=0))

        # update variance counter
        if binfo.mask is None or log_type == "grad":
            variance_counter[module_name][log_type] += batch_size
        else:
            variance_counter[module_name][log_type] += binfo.mask.sum().item()



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional

import torch
from einops import einsum, rearrange, reduce

from logix.state import LogIXState
from logix.statistic.utils import make_2d
from logix.utils import nested_dict


def precondition_kfac(
    src: Dict[str, Dict[str, torch.Tensor]],
    state: LogIXState,
    damping: Optional[float] = None,
) -> Dict[str, Dict[str, torch.Tensor]]:
    preconditioned = nested_dict()
    cov_eigval, cov_eigvec = state.get_covariance_svd_state()
    for module_name in src.keys():
        src_grad = src[module_name]["grad"]
        device = src_grad.device

        module_eigvec = cov_eigvec[module_name]
        fwd_eigvec = module_eigvec["forward"].to(device=device)
        bwd_eigvec = module_eigvec["backward"].to(device=device)

        # Reconstruct the full eigenvalue matrix with the damping factor added
        module_eigval = cov_eigval[module_name]
        if isinstance(module_eigval, torch.Tensor):
            full_eigval = module_eigval.to(device=device)
        else:
            assert "forward" in module_eigval and "backward" in module_eigval
            fwd_eigval = module_eigval["forward"]
            bwd_eigval = module_eigval["backward"]
            full_eigval = torch.outer(bwd_eigval, fwd_eigval).to(device=device)
        if damping is None:
            damping = 0.1 * torch.mean(full_eigval)
        full_eigval += damping

        # Precondition the gradient using eigenvectors and eigenvalues
        rotated_grad = einsum(
            bwd_eigvec.t(),
            src_grad,
            fwd_eigvec,
            "a b, batch b c, c d -> batch a d",
        )
        prec_rotated_grad = rotated_grad / full_eigval
        preconditioned[module_name]["grad"] = einsum(
            bwd_eigvec,
            prec_rotated_grad,
            fwd_eigvec.t(),
            "a b, batch b c, c d -> batch a d",
        )

    return preconditioned


def precondition_raw(
    src: Dict[str, Dict[str, torch.Tensor]],
    state: LogIXState,
    damping: Optional[float] = None,
) -> Dict[str, Dict[str, torch.Tensor]]:
    preconditioned = nested_dict()
    cov_inverse = state.get_covariance_inverse_state(damping=damping)
    for module_name in src.keys():
        device = src[module_name]["grad"].device
        grad_cov_inverse = cov_inverse[module_name]["grad"].to(device=device)
        original_shape = src[module_name]["grad"].shape
        preconditioned[module_name]["grad"] = (
            make_2d(src[module_name]["grad"], None, "grad") @ grad_cov_inverse
        ).reshape(original_shape)

    return preconditioned


def cross_dot_product(src: torch.Tensor, tgt: torch.Tensor) -> torch.Tensor:
    assert src.shape[1:] == tgt.shape[1:]
    src_expanded = rearrange(src, "n ... -> n 1 ...")
    tgt_expanded = rearrange(tgt, "m ... -> 1 m ...")
    dot_product_result = reduce(
        src_expanded * tgt_expanded,
        "n m ... -> n m",
        "sum",
    )

    return dot_product_result


def merge_influence_results(
    result_all: Dict[str, Dict[str, torch.Tensor]],
    result: Dict[str, Dict[str, torch.Tensor]],
    axis: str = "tgt",
) -> None:
    assert axis in ["src", "tgt"], f"Unsupported axis {axis}."

    # If merged result is empty, just copy the result and return
    if not result_all:
        result_all.update(result)
        return

    dim = int(axis == "tgt")
    id_key = f"{axis}_ids"

    result_all[id_key].extend(result[id_key])
    if isinstance(result["influence"], dict):
        for key in result_all["influence"].keys():
            result_all["influence"][key] = torch.cat(
                [result_all["influence"][key], result["influence"][key]], dim=dim
            )
    else:
        assert isinstance(result["influence"], torch.Tensor)
        result_all["influence"] = torch.cat(
            [result_all["influence"], result["influence"]], dim=dim
        )



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Optional

from lightning.pytorch.callbacks import Callback

from logix import LogIX, LogIXScheduler
from logix.lightning.arguments import LogIXArguments
from logix.utils import DataIDGenerator


class LogIXCallback(Callback):
    def __init__(
        self,
        logix: LogIX,
        logix_scheduler: LogIXScheduler,
        logix_args: LogIXArguments,
        data_id_extractor: Optional[Callable] = None,
        mask_extractor: Optional[Callable] = None,
    ) -> None:
        self.logix = logix
        self.scheduler = logix_scheduler
        self.args = logix_args

        self.data_id_extractor = data_id_extractor or DataIDGenerator()
        self.mask_extractor = mask_extractor

        self.counter = 0  # maybe use for gradient accumulation later

    def on_train_epoch_start(self, trainer, pl_module):
        if self.args.mode == "log":
            next(self.scheduler)
        else:
            self.logix.setup({"grad": ["log"]})
            self.logix.eval()

    def on_train_epoch_end(self, trainer, pl_module):
        if self.args.mode == "log":
            self.logix.finalize()

        max_epochs = len(self.scheduler) if self.args.mode == "log" else 1
        if trainer.current_epoch >= max_epochs - 1:
            trainer.should_stop = True

    def on_train_batch_start(self, trainer, pl_module, batch, batch_idx):
        data_id = self.data_id_extractor(batch)
        mask = None if self.mask_extractor is None else self.mask_extractor(batch)

        self.logix.start(data_id=data_id, mask=mask)

    def on_train_batch_end(self, trainer, pl_module, outputs, batch, batch_idx):
        self.logix.end()

        if self.args.mode == "influence":
            batch_log = self.logix.get_log()
            log_dataloader = self.logix.build_log_dataloader(
                num_workers=self.args.log_num_workers,
                batch_size=self.args.log_batch_size,
            )
            self.logix.influence.compute_influence_all(
                batch_log,
                log_dataloader,
                mode=self.args.influence_mode,
                damping=self.args.influence_damping,
                save=True,
            )
        elif self.args.mode == "self_influence":
            batch_log = self.logix.get_log()
            log_dataloader = self.logix.build_log_dataloader(
                num_workers=self.args.log_num_workers,
                batch_size=self.args.log_batch_size,
            )
            self.logix.influence.compute_self_influence(
                batch_log, damping=self.args.influence_damping
            )

    def on_before_optimizer_step(self, trainer, pl_module, optimizer):
        optimizer.zero_grad()



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Dict, List, Optional, Tuple

import torch
from einops import reduce
from tqdm import tqdm

from logix.analysis.influence_function_utils import (
    cross_dot_product,
    merge_influence_results,
    precondition_kfac,
    precondition_raw,
)
from logix.state import LogIXState
from logix.utils import flatten_log, get_logger, synchronize_device, unflatten_log


class InfluenceFunction:
    def __init__(self, state: LogIXState):
        # state
        self._state = state

        self.influence_scores = {}
        self.self_influence_scores = {}

    def get_influence_scores(self):
        return self.influence_scores

    def get_self_influence_scores(self):
        return self.self_influence_scores

    @torch.no_grad()
    def precondition(
        self,
        src_log: Dict[str, Dict[str, torch.Tensor]],
        damping: Optional[float] = None,
        hessian: Optional[str] = "auto",
    ) -> Tuple[List[str], Dict[str, Dict[str, torch.Tensor]]]:
        """
        Precondition gradients using the Hessian.

        Args:
            src_log (Dict[str, Dict[str, torch.Tensor]]): Log of source gradients
            damping (Optional[float], optional): Damping parameter for preconditioning. Defaults to None.
        """
        assert hessian in ["auto", "kfac", "raw"], f"Invalid hessian {hessian}"

        src_ids, src = src_log
        cov_state = self._state.get_covariance_state()
        if len(set(src.keys()) - set(cov_state.keys())) != 0:
            get_logger().warning(
                "Not all covariances have been computed. No preconditioning applied.\n"
            )
            return src_log

        precondition_fn = precondition_kfac
        if hessian == "raw" or (
            hessian == "auto" and "grad" in cov_state[list(src.keys())[0]]
        ):
            precondition_fn = precondition_raw
        preconditioned_grad = precondition_fn(
            src=src, state=self._state, damping=damping
        )

        return (src_ids, preconditioned_grad)

    @torch.no_grad()
    def compute_influence(
        self,
        src_log: Tuple[str, Dict[str, Dict[str, torch.Tensor]]],
        tgt_log: Tuple[str, Dict[str, Dict[str, torch.Tensor]]],
        mode: Optional[str] = "dot",
        precondition: Optional[bool] = True,
        hessian: Optional[str] = "auto",
        influence_groups: Optional[List[str]] = None,
        damping: Optional[float] = None,
    ):
        """
        Compute influence scores between two gradient dictionaries.

        Args:
            src_log (Dict[str, Dict[str, torch.Tensor]]): Log of source gradients
            tgt_log (Dict[str, Dict[str, torch.Tensor]]): Log of target gradients
            mode (Optional[str], optional): Influence function mode. Defaults to "dot".
            precondition (Optional[bool], optional): Whether to precondition the gradients. Defaults to True.
            damping (Optional[float], optional): Damping parameter for preconditioning. Defaults to None.
        """
        assert mode in ["dot", "l2", "cosine"], f"Invalid mode: {mode}"

        result = {}
        if precondition:
            src_log = self.precondition(
                src_log=src_log, damping=damping, hessian=hessian
            )

        src_ids, src = src_log
        tgt_ids, tgt = tgt_log

        # Initialize influence scores
        total_influence = {"total": 0}
        for influence_group in influence_groups or []:
            total_influence[influence_group] = 0

        # Compute influence scores. By default, we should compute the basic influence
        # scores, which is essentially the inner product between the source and target
        # gradients. If mode is cosine, we should normalize the influence score by the
        # L2 norm of the target gardients. If mode is l2, we should subtract the L2
        # norm of the target gradients.
        if not isinstance(tgt, dict):
            assert isinstance(tgt, torch.Tensor)
            src = flatten_log(
                log=src, path=self._state.get_state("model_module")["path"]
            )
            tgt = tgt.to(device=src.device)
            total_influence["total"] += cross_dot_product(src, tgt)
        else:
            synchronize_device(src, tgt)
            for module_name in src.keys():
                module_influence = cross_dot_product(
                    src[module_name]["grad"], tgt[module_name]["grad"]
                )
                total_influence["total"] += module_influence
                if influence_groups is not None:
                    groups = [g for g in influence_groups if g in module_name]
                    for group in groups:
                        total_influence[group] += module_influence

        if mode == "cosine":
            tgt_norm = self.compute_self_influence(
                tgt_log,
                precondition=True,
                hessian=hessian,
                influence_groups=influence_groups,
                damping=damping,
            ).pop("influence")
            for key in total_influence.keys():
                tgt_norm_key = tgt_norm if influence_groups is None else tgt_norm[key]
                total_influence[key] /= torch.sqrt(tgt_norm_key.unsqueeze(0))
        elif mode == "l2":
            tgt_norm = self.compute_self_influence(
                tgt_log,
                precondition=True,
                hessian=hessian,
                influence_groups=influence_groups,
                damping=damping,
            ).pop("influence")
            for key in total_influence.keys():
                tgt_norm_key = tgt_norm if influence_groups is None else tgt_norm[key]
                total_influence[key] -= 0.5 * tgt_norm_key.unsqueeze(0)

        # Move influence scores to CPU to save memory
        for key, value in total_influence.items():
            assert value.shape[0] == len(src_ids)
            assert value.shape[1] == len(tgt_ids)
            total_influence[key] = value.cpu()

        result["src_ids"] = list(src_ids)
        result["tgt_ids"] = list(tgt_ids)
        result["influence"] = (
            total_influence.pop("total")
            if influence_groups is None
            else total_influence
        )

        return result

    @torch.no_grad()
    def compute_self_influence(
        self,
        src_log: Tuple[str, Dict[str, Dict[str, torch.Tensor]]],
        precondition: Optional[bool] = True,
        hessian: Optional[str] = "auto",
        influence_groups: Optional[List[str]] = None,
        damping: Optional[float] = None,
    ):
        """
        Compute self-influence scores. This can be used for uncertainty estimation.

        Args:
            src_log (Dict[str, Dict[str, torch.Tensor]]): Log of source gradients
            precondition (Optional[bool], optional): Whether to precondition the gradients. Defaults to True.
            damping (Optional[float], optional): Damping parameter for preconditioning. Defaults to None.
        """
        result = {}

        src_ids, src = src_log
        if not isinstance(src, dict):
            assert isinstance(src, torch.Tensor)
            src = unflatten_log(
                log=src, path=self._state.get_state("model_module")["path"]
            )

        tgt = src
        if precondition:
            tgt = self.precondition(src_log, hessian=hessian, damping=damping)[1]

        # Initialize influence scores
        total_influence = {"total": 0}
        for influence_group in influence_groups or []:
            total_influence[influence_group] = 0

        # Compute self-influence scores
        for module_name in src.keys():
            src_module = src[module_name]["grad"]
            tgt_module = tgt[module_name]["grad"] if tgt is not None else src_module
            module_influence = reduce(
                src_module * tgt_module, "n a b -> n", "sum"
            ).reshape(-1)
            total_influence["total"] += module_influence
            if influence_groups is not None:
                groups = [g for g in influence_groups if g in module_name]
                for group in groups:
                    total_influence[group] += module_influence

        # Move influence scores to CPU to save memory
        for key, value in total_influence.items():
            assert len(value) == len(src_ids)
            total_influence[key] = value.cpu()

        result["src_ids"] = src_ids
        result["influence"] = (
            total_influence.pop("total")
            if influence_groups is None
            else total_influence
        )

        return result

    def compute_influence_all(
        self,
        src_log: Tuple[str, Dict[str, Dict[str, torch.Tensor]]],
        loader: torch.utils.data.DataLoader,
        mode: Optional[str] = "dot",
        precondition: Optional[bool] = True,
        save: Optional[bool] = False,
        hessian: Optional[str] = "auto",
        influence_groups: Optional[List[str]] = None,
        damping: Optional[float] = None,
    ):
        """
        Compute influence scores against all traininig data in the log. This can be used
        for training data attribution.

        Args:
            src_log (Dict[str, Dict[str, torch.Tensor]]): Log of source gradients
            loader (torch.utils.data.DataLoader): DataLoader of train data
            mode (Optional[str], optional): Influence function mode. Defaults to "dot".
            precondition (Optional[bool], optional): Whether to precondition the gradients. Defaults to True.
            damping (Optional[float], optional): Damping parameter for preconditioning. Defaults to None.
        """
        if precondition:
            src_log = self.precondition(src_log, hessian=hessian, damping=damping)

        result_all = {}
        for tgt_log in tqdm(loader, desc="Compute IF"):
            result = self.compute_influence(
                src_log=src_log,
                tgt_log=tgt_log,
                mode=mode,
                precondition=False,
                hessian=hessian,
                influence_groups=influence_groups,
                damping=damping,
            )
            merge_influence_results(result_all, result, axis="tgt")

        if save:
            merge_influence_results(self.influence_scores, result_all, axis="src")

        return result_all



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from concurrent.futures import ThreadPoolExecutor

import torch

from logix.logging.mmap import MemoryMapHandler
from logix.utils import get_rank, nested_dict, to_numpy


class LogSaver:
    def __init__(self, config, state):
        self.log_dir = config.log_dir
        self.state = state
        self.model_module = self.state.get_state("model_module")
        self.file_prefix = f"log_rank_{get_rank()}_chunk_"

        self.max_worker = config.num_workers
        self.allow_async = True if self.max_worker > 1 else False

        self.flush_threshold = config.flush_threshold
        self.flush_count = 0

        self.buffer = nested_dict()
        self.buffer_size = 0

    def buffer_write(self, binfo):
        """
        Add log state on exit.
        """
        data_id = binfo.data_id
        log = binfo.log

        def _add(log, buffer, idx):
            for key, value in log.items():
                if isinstance(value, torch.Tensor):
                    numpy_value = to_numpy(value[idx])
                    buffer[key] = numpy_value
                    self.buffer_size += numpy_value.size
                    continue
                _add(value, buffer[key], idx)

        for idx, did in enumerate(data_id):
            _add(log, self.buffer[did], idx)

    def _flush_unsafe(self, log_dir, buffer, flush_count) -> str:
        """
        _flush_unsafe is thread unsafe flush of current buffer. No shared variable must be allowed.
        """
        filename = self.file_prefix + f"{flush_count}.mmap"
        buffer_list = [(k, v) for k, v in buffer]
        MemoryMapHandler.write(
            log_dir, filename, buffer_list, self.model_module["path"]
        )
        return filename

    def _flush_safe(self, log_dir) -> str:
        """
        _flush_safe is thread safe flush of current buffer.
        """
        buffer_copy = self.buffer.copy()
        flush_count_copy = self.flush_count
        self.flush_count += 1
        self.buffer_clear()
        with ThreadPoolExecutor(max_workers=self.max_worker) as executor:
            save_path = executor.submit(
                self._flush_unsafe, log_dir, buffer_copy, flush_count_copy
            )
        return save_path

    def _flush_serialized(self, log_dir) -> str:
        """
        _flush_serialized executes the flushing of the buffers in serialized manner.
        """
        if len(self.buffer) == 0:
            return log_dir
        buffer_list = [(k, v) for k, v in self.buffer.items()]
        MemoryMapHandler.write(
            log_dir,
            self.file_prefix + f"{self.flush_count}.mmap",
            buffer_list,
            self.model_module["path"],
            dtype="uint8",
        )

        self.flush_count += 1
        self.buffer_clear()
        del buffer_list
        return log_dir

    def flush(self) -> None:
        """
        For the DefaultHandler, there's no batch operation needed since each add operation writes to the file.
        This can be a placeholder or used for any finalization operations.
        """
        if 0 < self.flush_threshold < self.buffer_size:
            if self.allow_async:
                self._flush_safe(self.log_dir)
                return
            self._flush_serialized(self.log_dir)

    def finalize(self):
        """
        Dump everything in the buffer to disk when `logix.finalize()` is called.
        """
        self._flush_serialized(self.log_dir)

    def buffer_clear(self):
        """
        Clears the buffer.
        """
        self.buffer.clear()
        self.buffer_size = 0



from logix.logix import LogIX
from logix.scheduler import LogIXScheduler
from logix.utils import get_logger

__version__ = "0.1.1"

_LOGIX_INSTANCE = None


def init(project: str, config: str = "./config.yaml"):
    """Initialize LogIX.

    Args:
        project (str): The name of the project.
        config (dict): The configuration dictionary.
    """
    global _LOGIX_INSTANCE

    if _LOGIX_INSTANCE is not None:
        get_logger().warning(
            "LogIX is already initialized. If you want to initialize "
            + "additional LogIX instances, please use logix.LogIX instead."
        )
        return

    run = LogIX(project, config)

    _LOGIX_INSTANCE = run

    return run


def add_analysis(*args, **kwargs):
    if _LOGIX_INSTANCE is None:
        raise RuntimeError(
            "LogIX is not initialized. You must call logix.init() first."
        )
    return _LOGIX_INSTANCE.add_analysis(*args, **kwargs)


def add_lora(*args, **kwargs):
    if _LOGIX_INSTANCE is None:
        raise RuntimeError(
            "LogIX is not initialized. You must call logix.init() first."
        )
    return _LOGIX_INSTANCE.add_lora(*args, **kwargs)


def build_log_dataloader(*args, **kwargs):
    if _LOGIX_INSTANCE is None:
        raise RuntimeError(
            "LogIX is not initialized. You must call logix.init() first."
        )
    return _LOGIX_INSTANCE.build_log_dataloader(*args, **kwargs)


def clear(*args, **kwargs):
    if _LOGIX_INSTANCE is None:
        raise RuntimeError(
            "LogIX is not initialized. You must call logix.init() first."
        )
    return _LOGIX_INSTANCE.clear(*args, **kwargs)


def eval(*args, **kwargs):
    if _LOGIX_INSTANCE is None:
        raise RuntimeError(
            "LogIX is not initialized. You must call logix.init() first."
        )
    return _LOGIX_INSTANCE.eval(*args, **kwargs)


def finalize(*args, **kwargs):
    if _LOGIX_INSTANCE is None:
        raise RuntimeError(
            "LogIX is not initialized. You must call logix.init() first."
        )
    return _LOGIX_INSTANCE.finalize(*args, **kwargs)


def get_log(*args, **kwargs):
    if _LOGIX_INSTANCE is None:
        raise RuntimeError(
            "LogIX is not initialized. You must call logix.init() first."
        )
    return _LOGIX_INSTANCE.get_log(*args, **kwargs)


def initialize_from_log(*args, **kwargs):
    if _LOGIX_INSTANCE is None:
        raise RuntimeError(
            "LogIX is not initialized. You must call logix.init() first."
        )
    return _LOGIX_INSTANCE.initialize_from_log(*args, **kwargs)


def log(*args, **kwargs):
    if _LOGIX_INSTANCE is None:
        raise RuntimeError(
            "LogIX is not initialized. You must call logix.init() first."
        )
    return _LOGIX_INSTANCE.log(*args, **kwargs)


def setup(*args, **kwargs):
    if _LOGIX_INSTANCE is None:
        raise RuntimeError(
            "LogIX is not initialized. You must call logix.init() first."
        )
    return _LOGIX_INSTANCE.setup(*args, **kwargs)


def watch(*args, **kwargs):
    if _LOGIX_INSTANCE is None:
        raise RuntimeError(
            "LogIX is not initialized. You must call logix.init() first."
        )
    return _LOGIX_INSTANCE.watch(*args, **kwargs)


def watch_activation(*args, **kwargs):
    if _LOGIX_INSTANCE is None:
        raise RuntimeError(
            "LogIX is not initialized. You must call logix.init() first."
        )
    return _LOGIX_INSTANCE.watch_activation(*args, **kwargs)



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import re
from collections import OrderedDict
from functools import reduce
from typing import List

import numpy as np
import torch
from torch.utils.data import default_collate

from logix.logging.mmap import MemoryMapHandler


def extract_rank_and_chunk(filename):
    """
    Extracts the rank and chunk index from the filename.

    Args:
        filename (str): Filename to extract rank and chunk index from.

    Returns:
        Tuple[int, int]: Tuple containing the rank and chunk index.
    """
    match = re.search(r"rank_(\d+)_chunk_(\d+)", filename)
    return int(match.group(1)), int(match.group(2))


def find_chunk_indices(path) -> List:
    """
    Finds and returns the sorted list of chunk indices based on the filenames in the input path.

    Args:
        path (str): The path to search for chunk files.

    Returns:
        List[int]: Sorted list of chunk indices.
    """
    chunk_indices = []
    for filename in os.listdir(path):
        if filename.endswith(".mmap") and filename.startswith("log_"):
            chunk_index = filename.rstrip(".mmap").strip("log_")
            chunk_indices.append(chunk_index)

    return sorted(chunk_indices, key=extract_rank_and_chunk)


def get_mmap_data(path, mmap_filename, dtype="uint8") -> List:
    """
    Adds memory-mapped files for the given mmap file.

    Args:
        path (str): Path to the directory containing the mmap file.
        mmap_filename (str): Filename of the mmap file.

    Returns:
       List: A list of memory maps.
    """
    with MemoryMapHandler.read(path, mmap_filename, dtype) as mm:
        return mm


def get_mmap_metadata(
    data_id_to_chunk, path, metadata_filename, chunk_index
) -> OrderedDict:
    metadata = MemoryMapHandler.read_metafile(path, metadata_filename)
    # Update the mapping from data_id to chunk
    for entry in metadata:
        if entry["data_id"] in data_id_to_chunk:
            logging.warning(f"duplicated data_id detected: {entry['data_id']}")
            continue
        data_id_to_chunk[entry["data_id"]] = (chunk_index, entry)
    return data_id_to_chunk


def collate_nested_dicts(batch):
    # `batch` is a list of tuples, each tuple is (data_id, nested_dict)
    batched_data_ids = [data_id for data_id, _ in batch]

    # Initialize the batched_nested_dicts by deep copying the first nested_dict structure
    # Replace all tensors with lists to hold tensors from all items in the batch
    first_nested_dict = batch[0][1]
    batched_nested_dicts = {
        k: _init_collate_structure(v) for k, v in first_nested_dict.items()
    }

    # Now iterate through all items and populate the batched_nested_dicts
    for _, nested_dict in batch:
        _merge_dicts(batched_nested_dicts, nested_dict)

    # Finally, we should collate the lists of tensors into batched tensors
    _collate_tensors_in_structure(batched_nested_dicts)

    return batched_data_ids, batched_nested_dicts


def get_entry_metadata(entries):
    blocks = []
    dtype = None
    for entry in entries:
        blocks.append(reduce(lambda x, y: x * y, entry["shape"]))
        dtype = np.dtype(entry["dtype"])
    blocksize = reduce(lambda x, y: x + y, blocks)
    return blocksize, dtype


def get_flatten_item(mmap, offset, block_size, dtype="float32"):
    array = np.ndarray(
        block_size,
        dtype,
        buffer=mmap,
        offset=offset,
        order="C",
    )
    return torch.from_numpy(array).clone()


def unflatten_tensor(flat_tensor, shape, start):
    num_elements = reduce(lambda x, y: x * y, shape)
    end = start + num_elements
    unflattened_tensor = flat_tensor[start:end].view(*shape)
    return unflattened_tensor, end


def _init_collate_structure(nested_dict):
    # Initialize the collate structure based on the first item
    if isinstance(nested_dict, dict):
        return {k: _init_collate_structure(v) for k, v in nested_dict.items()}
    else:
        return []


def _merge_dicts(accumulator, new_data):
    # Merge new_data into the accumulator recursively
    for key, value in new_data.items():
        if isinstance(value, dict):
            # Recursive call if the value is a dictionary
            _merge_dicts(accumulator[key], value)
        else:
            # Assume the value is a tensor, append it to the list in accumulator
            accumulator[key].append(value)


def _collate_tensors_in_structure(nested_dict):
    # Collate all lists of tensors within the nested structure
    for key, value in nested_dict.items():
        if isinstance(value, dict):
            # Recursive call if the value is a dictionary
            _collate_tensors_in_structure(value)
        else:
            # Stack all tensors in the list along a new batch dimension
            nested_dict[key] = default_collate(value)



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Optional

import torch.nn as nn

from logix.config import LoRAConfig
from logix.lora.modules import LoraConv2d, LoraEmbedding, LoraLinear
from logix.lora.utils import _get_submodules, find_parameter_sharing_group
from logix.state import LogIXState
from logix.utils import get_logger, module_check


class LoRAHandler:
    """
    Transforms a model into a Lora model.
    """

    _SUPPORTED_MODULES = {nn.Linear, nn.Conv1d, nn.Conv2d}

    def __init__(self, config: LoRAConfig, state: LogIXState):
        self._state = state

        self.init_strategy = config.init
        self.rank = config.rank
        self.parameter_sharing = config.parameter_sharing
        self.parameter_sharing_groups = config.parameter_sharing_groups

    def add_lora(
        self,
        model: nn.Module,
        type_filter: Optional[List[nn.Module]],
        name_filter: Optional[List[str]],
    ):
        """
        Add LoRA modules to a model.
        """
        covariance_state = self._state.get_covariance_state()
        if self.init_strategy == "pca" and len(covariance_state) == 0:
            get_logger().warning(
                "Hessian state not provided. Using random initialization instead."
            )
            self.init_strategy = "random"

        shared_modules = {}
        device = next(model.parameters()).device
        for name, module in model.named_modules():
            if not module_check(
                module=module,
                module_name=name,
                supported_modules=self._SUPPORTED_MODULES,
                type_filter=type_filter,
                name_filter=name_filter,
            ):
                continue

            # Add Lora to filtered modules
            lora_cls = None
            if isinstance(module, nn.Linear):
                lora_cls = LoraLinear
            elif isinstance(module, nn.Conv1d):
                raise NotImplementedError
            elif isinstance(module, nn.Conv2d):
                lora_cls = LoraConv2d
            elif isinstance(module, nn.Embedding):
                lora_cls = LoraEmbedding

            psg = find_parameter_sharing_group(name, self.parameter_sharing_groups)
            if self.parameter_sharing and psg not in shared_modules:
                if isinstance(module, nn.Linear):
                    shared_module = nn.Linear(self.rank, self.rank, bias=False)
                elif isinstance(module, nn.Conv1d):
                    shared_module = nn.Conv1d(
                        self.rank, self.rank, kernel_size=1, bias=False
                    )
                elif isinstance(module, nn.Conv2d):
                    shared_module = nn.Conv2d(
                        self.rank, self.rank, kernel_size=1, bias=False
                    )
                shared_modules[psg] = shared_module

            lora_module = lora_cls(self.rank, module, shared_modules.get(psg, None))
            if self.init_strategy == "pca":
                lora_module.pca_init_weight(covariance_state[name])
            lora_module.to(device)

            parent, _, target_name = _get_submodules(model, name)
            setattr(parent, target_name, lora_module)



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from copy import deepcopy
from functools import reduce
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import torch
import torch.nn as nn

from logix.analysis import InfluenceFunction
from logix.batch_info import BatchInfo
from logix.config import Config, LoggingConfig, LoRAConfig, init_config_from_yaml
from logix.logging import HookLogger
from logix.logging.log_loader import LogDataset
from logix.logging.log_loader_utils import collate_nested_dicts
from logix.lora import LoRAHandler
from logix.lora.utils import is_lora
from logix.state import LogIXState
from logix.utils import (
    get_logger,
    get_rank,
    get_repr_dim,
    get_world_size,
    module_check,
    print_tracked_modules,
)


class LogIX:
    """
    LogIX is a front-end interface for logging and analyzing neural networks.
    Using (PyTorch) hooks, it tracks, saves, and computes statistics for activations,
    gradients, and other tensors with its logger. Once logging is finished, it provides
    an interface for computing influence scores and other analysis.

    Args:
        project (str):
            The name or identifier of the project. This is used for organizing
            logs and analysis results.
        config (str, optional):
            The path to the YAML configuration file. This file contains settings
            for logging, analysis, and other features. Defaults to an empty string,
            which means default settings are used.
    """

    _SUPPORTED_MODULES = {nn.Linear, nn.Conv1d, nn.Conv2d}

    def __init__(
        self,
        project: str,
        config: Optional[str] = None,
        logging_config: Optional[LoggingConfig] = None,
    ) -> None:
        self.project: str = project

        self.model: Optional[nn.Module] = None

        # Config
        self.config: Config = init_config_from_yaml(
            project=project, logix_config=config
        )
        if logging_config is not None:
            self.set_logging_config(logging_config)
        self.log_dir: str = self.config.log_dir

        # LogIX state
        self.state: LogIXState = LogIXState()
        self.binfo: BatchInfo = BatchInfo()
        self._save: bool = False
        self._save_batch: bool = False

        # Initialize logger
        self.logger: HookLogger = HookLogger(
            config=self.get_logging_config(), state=self.state, binfo=self.binfo
        )

        # Log data
        self.log_dataset: Optional[torch.utils.data.Dataset] = None
        self.log_dataloader: Optional[torch.utils.data.DataLoader] = None

        # Analysis
        self.influence: InfluenceFunction = InfluenceFunction(state=self.state)

        # Misc
        self.type_filter: Optional[List[nn.Module]] = None
        self.name_filter: Optional[List[str]] = None

    def watch(
        self,
        model: nn.Module,
        type_filter: Optional[List[nn.Module]] = None,
        name_filter: Optional[List[str]] = None,
    ) -> None:
        """
        Sets up modules in the model to be watched based on optional type and name filters.
        Hooks will be added to the watched modules to track their forward activations,
        backward error signals, and gradient, and compute various statistics (e.g. mean,
        variance, covariance) for each of these.

        Args:
            model (nn.Module):
                The neural network model to be watched.
            type_filter (List[nn.Module], optional):
                A list of module types to be watched. If None, all supported module types
                are watched.
            name_filter (List[str], optional):
                A list of module names to be watched. If None, all modules are considered.
        """
        self.model = model
        if get_world_size() > 1 and hasattr(self.model, "module"):
            self.model = self.model.module
        self.type_filter = type_filter or self.type_filter
        self.name_filter = name_filter or self.name_filter

        _is_lora = is_lora(self.model)

        for name, module in self.model.named_modules():
            if module_check(
                module=module,
                module_name=name,
                supported_modules=self._SUPPORTED_MODULES,
                type_filter=self.type_filter,
                name_filter=self.name_filter,
                is_lora=_is_lora,
            ):
                self.logger.add_module(name, module)
            elif len(list(module.children())) == 0:
                # disable gradient compute for non-tracked modules. This will save a
                # significant amount of GPU memory and compute for large models
                for p in module.parameters():
                    p.requires_grad = False

        module_path, repr_dim = get_repr_dim(self.logger.modules_to_name)
        print_tracked_modules(reduce(lambda x, y: x + y, repr_dim))
        self.state.set_state("model_module", path=module_path, path_dim=repr_dim)

    def watch_activation(self, tensor_dict: Dict[str, torch.Tensor]) -> None:
        """
        Sets up tensors to be watched. This is useful for logging custom tensors or
        activations that are not directly associated with model modules.

        Args:
            tensor_dict (Dict[str, torch.Tensor]):
                A dictionary where keys are tensor names and values are the tensors
                themselves.
        """
        self.logger.register_all_tensor_hooks(tensor_dict)

    def add_lora(
        self,
        watch: bool = True,
        clear: bool = True,
        type_filter: Optional[List[nn.Module]] = None,
        name_filter: Optional[List[str]] = None,
        lora_path: Optional[str] = None,
        lora_config: Optional[LoRAConfig] = None,
    ) -> None:
        """
        Adds an LoRA variant for gradient compression. Note that the added LoRA module
        is different from the original LoRA module in that it consists of three parts:
        encoder, bottleneck, and decoder. The encoder and decoder are initialized with
        random weights, while the bottleneck is initialized to be zero. In doing so,
        the encoder and decoder are repsectively used to compress forward and backward
        signals, while the bottleneck is used to extract the compressed gradient.
        Mathematically, this corresponds to random Kronecker product compression.

        Args:
            model (Optional[nn.Module]):
                The model to apply LoRA to. If None, the model already set with watch
                is used.
            watch (bool, optional):
                Whether to watch the model after adding LoRA. Defaults to `True`.
            clear (bool, optional):
                Whether to clear the internal states after adding LoRA. Defaults to
                `True`.
            type_filter (Optional[List[nn.Module]], optional):
                A list of module types to be watched.
            name_filter (Optional[List[str]], optional):
                A list of module names to be watched.
            lora_path (Optional[str], optional):
                The path to the LoRA state file. If None, the LoRA state is not loaded.
            lora_config (Optional[LoRAConfig], optional): LoRA configuration.
        """
        if lora_config is not None:
            self.set_lora_config(lora_config)

        if not hasattr(self, "lora_handler"):
            self.lora_handler = LoRAHandler(self.get_lora_config(), self.state)

        self.lora_handler.add_lora(
            model=self.model,
            type_filter=type_filter or self.type_filter,
            name_filter=name_filter or self.name_filter,
        )

        # If lora_path is not none, directly load lora weights from this path
        if lora_path is not None:
            lora_dir = os.path.join(os.path.join(lora_path, "lora"))
            lora_state = torch.load(os.path.join(lora_dir, "lora_state_dict.pt"))
            for name in lora_state:
                assert name in self.model.state_dict(), f"{name} not in model!"
            self.model.load_state_dict(lora_state, strict=False)

        # Save lora state dict
        if get_rank() == 0:
            self.save_lora()

        # Clear state and logger
        if clear or watch:
            msg = "LogIX will clear the previous Hessian, Storage, and Logging "
            msg += "handlers after adding LoRA for gradient compression.\n"
            get_logger().info(msg)
            self.clear()

        # (Re-)watch lora-added model
        if watch:
            self.watch(self.model)

    def log(self, data_id: Iterable[Any], mask: Optional[torch.Tensor] = None) -> None:
        """
        Logs the data. This is an experimental feature for now.

        Args:
            data_id (str):
                A unique identifier associated with the data for the logging session.
            mask (Optional[torch.Tensor], optional):
                An optional attention mask used in Transformer models.
        """
        self.logger.log(data_id=data_id, mask=mask)

    def __call__(
        self,
        data_id: Iterable[Any],
        mask: Optional[torch.Tensor] = None,
        save: bool = False,
    ):
        """
        Args:
            data_id: A unique identifier associated with the data for the logging session.
            mask (torch.Tensor, optional): Mask for the data.

        Returns:
            self: Returns the instance of the LogIX object.
        """
        self.binfo.clear()
        self.binfo.data_id = data_id
        self.binfo.mask = mask

        self._save_batch = save

        return self

    def __enter__(self):
        """
        Sets up the context manager.

        This method is automatically called when the `with` statement is used with an `LogIX` object.
        It sets up the logging environment based on the provided parameters.
        """
        self.logger.clear(hook=True, module=False, buffer=False)
        self.logger.register_all_module_hooks()

        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """
        Clears the internal states and removes all the hooks.

        This method is essential for ensuring that there are no lingering hooks that could
        interfere with further operations on the model or with future logging sessions.
        """
        self.logger.update(save=self._save_batch or self._save)

    def start(
        self, data_id: Iterable[Any], mask: Optional[torch.Tensor] = None
    ) -> None:
        """
        This is another programming interface for logging. Instead of using the context manager, we also
        allow users to manually specify `start` and `end` points for logging. In the `start` point,
        users should specify `data_id` and optionally `mask`.

        Args:
            data_id: A unique identifier associated with the data for the logging session.
            mask (torch.Tensor, optional): Mask for the data.
        """
        # Set up batch information
        self.binfo.clear()
        self.binfo.data_id = data_id
        self.binfo.mask = mask

        # Set up logger
        self.logger.clear(hook=True, module=False, buffer=False)
        self.logger.register_all_module_hooks()

    def end(self, save: bool = False) -> None:
        """
        This is another programming interface for logging. Instead of using the context manager, we also
        allow users to manually specify "start" and "end" points for logging.
        """
        self.logger.update(save=save or self._save)

    def build_log_dataset(self, flatten: bool = False) -> torch.utils.data.Dataset:
        """
        Constructs the log dataset from the stored logs. This dataset can then be used
        for analysis or visualization.

        Args:
            flatten (bool, optional): Whether to flatten the nested dictionary structure. Defaults to False.

        Returns:
            LogDataset:
                An instance of LogDataset containing the logged data.
        """
        if self.log_dataset is None:
            self.log_dataset = LogDataset(log_dir=self.log_dir, flatten=flatten)
        return self.log_dataset

    def build_log_dataloader(
        self,
        batch_size: int = 16,
        num_workers: int = 0,
        pin_memory: bool = False,
        flatten: bool = False,
    ) -> torch.utils.data.DataLoader:
        """
        Constructs a DataLoader for the log dataset. This is useful for batch processing
        of logged data during analysis. It also follows PyTorch DataLoader conventions.

        Args:
            batch_size (int, optional): The batch size for the DataLoader. Defaults to 16.
            num_workers (int, optional): The number of workers for the DataLoader. Defaults to 0.
            pin_memory (bool, optional): Whether to pin memory for the DataLoader. Defaults to False.
            flatten (bool, optional): Whether to flatten the nested dictionary structure. Defaults to False.

        Return:
            DataLoader:
                A DataLoader instance for the log dataset.
        """
        if self.log_dataloader is None:
            log_dataset = self.build_log_dataset(flatten=flatten)
            collate_fn = None if flatten else collate_nested_dicts
            self.log_dataloader = torch.utils.data.DataLoader(
                log_dataset,
                batch_size=batch_size,
                num_workers=num_workers,
                pin_memory=pin_memory,
                shuffle=False,
                collate_fn=collate_fn,
            )
        return self.log_dataloader

    def get_log(
        self, copy: bool = False
    ) -> Tuple[str, Dict[str, Dict[str, torch.Tensor]]]:
        """
        Returns the current log, including data identifiers and logged information.

        Args:
           copy (bool, optional): Whether to return a copy of the log. Defaults to False.

        Returns:
            dict: The current log.
        """
        log = (self.binfo.data_id, self.binfo.log)
        return log if not copy else deepcopy(log)

    def compute_influence(
        self,
        src_log: Tuple[str, Dict[str, Dict[str, torch.Tensor]]],
        tgt_log: Tuple[str, Dict[str, Dict[str, torch.Tensor]]],
        mode: Optional[str] = "dot",
        precondition: Optional[bool] = True,
        hessian: Optional[str] = "auto",
        influence_groups: Optional[List[str]] = None,
        damping: Optional[float] = None,
    ) -> Dict[str, Union[List[str], torch.Tensor, Dict[str, torch.Tensor]]]:
        """
        Front-end interface for computing influence scores. It calls the
        `compute_influence` method of the `InfluenceFunction` class.

        Args:
            src_log (Tuple[str, Dict[str, Dict[str, torch.Tensor]]]): Log of source gradients
            tgt_log (Tuple[str, Dict[str, Dict[str, torch.Tensor]]]): Log of target gradients
            mode (str, optional): Influence function mode. Defaults to "dot".
            precondition (bool, optional): Whether to precondition the gradients. Defaults to True.
            hessian (str, optional): Hessian computation mode. Defaults to "auto".
            influence_groups (List[str], optional): List of influence groups. Defaults to None.
            damping (float, optional): Damping factor. Defaults to None.
        """
        result = self.influence.compute_influence(
            src_log=src_log,
            tgt_log=tgt_log,
            mode=mode,
            precondition=precondition,
            hessian=hessian,
            influence_groups=influence_groups,
            damping=damping,
        )
        return result

    def compute_influence_all(
        self,
        src_log: Optional[Tuple[str, Dict[str, Dict[str, torch.Tensor]]]] = None,
        loader: Optional[torch.utils.data.DataLoader] = None,
        mode: Optional[str] = "dot",
        precondition: Optional[bool] = True,
        hessian: Optional[str] = "auto",
        influence_groups: Optional[List[str]] = None,
        damping: Optional[float] = None,
    ) -> Dict[str, Union[List[str], torch.Tensor, Dict[str, torch.Tensor]]]:
        """
        Front-end interface for computing influence scores against all train data in the log.
        It calls the `compute_influence_all` method of the `InfluenceFunction` class.

        Args:
            src_log (Tuple[str, Dict[str, Dict[str, torch.Tensor]]]): Log of source gradients
            loader (torch.utils.data.DataLoader): DataLoader of train data
            mode (str, optional): Influence function mode. Defaults to "dot".
            precondition (bool, optional): Whether to precondition the gradients. Defaults to True.
            hessian (str, optional): Hessian computation mode. Defaults to "auto".
            influence_groups (List[str], optional): List of influence groups. Defaults to None.
            damping (float, optional): Damping factor. Defaults to None.
        """
        src_log = src_log if src_log is not None else self.get_log()
        loader = loader if loader is not None else self.build_log_dataloader()

        result = self.influence.compute_influence_all(
            src_log=src_log,
            loader=loader,
            mode=mode,
            precondition=precondition,
            hessian=hessian,
            influence_groups=influence_groups,
            damping=damping,
        )
        return result

    def compute_self_influence(
        self,
        src_log: Optional[Tuple[str, Dict[str, Dict[str, torch.Tensor]]]] = None,
        precondition: Optional[bool] = True,
        hessian: Optional[str] = "auto",
        influence_groups: Optional[List[str]] = None,
        damping: Optional[float] = None,
    ) -> Dict[str, Union[List[str], torch.Tensor, Dict[str, torch.Tensor]]]:
        """
        Front-end interface for computing self-influence scores. It calls the
        `compute_self_influence` method of the `InfluenceFunction` class.

        Args:
            src_log (Tuple[str, Dict[str, Dict[str, torch.Tensor]]]): Log of source gradients
            precondition (bool, optional): Whether to precondition the gradients. Defaults to True.
            heissian (str, optional): Hessian computation mode. Defaults to "auto".
            influence_groups (List[str], optional): List of influence groups. Defaults to None.
            damping (float, optional): Damping factor. Defaults to None.
        """
        src_log = src_log if src_log is not None else self.get_log()
        result = self.influence.compute_self_influence(
            src_log=src_log,
            precondition=precondition,
            hessian=hessian,
            influence_groups=influence_groups,
            damping=damping,
        )
        return result

    def save_lora(self) -> None:
        """
        Save LoRA state to disk.
        """
        state_dict = self.model.state_dict()
        lora_state_dict = {
            name: param for name, param in state_dict.items() if "logix_lora" in name
        }
        if len(lora_state_dict) > 0:
            log_dir = os.path.join(self.log_dir, "lora")
            if not os.path.exists(log_dir) and get_rank() == 0:
                os.makedirs(log_dir)
            torch.save(lora_state_dict, os.path.join(log_dir, "lora_state_dict.pt"))

    def initialize_from_log(
        self, state_path: Optional[str] = None, lora_path: Optional[str] = None
    ) -> None:
        """
        Load all states from disk.

        Args:
            state_path (str, optional): Path to the state file.
            lora_path (str, optional): Path to the LoRA state file.
        """
        assert os.path.exists(self.log_dir), f"{self.log_dir} does not exist!"

        # Load config
        self.config.load_config(os.path.join(self.log_dir, "config.yaml"))

        # Load LoRA state
        lora_dir = os.path.join(lora_path or self.log_dir, "lora")
        if os.path.exists(lora_dir) and self.model is not None:
            if not is_lora(self.model):
                self.add_lora()
            lora_state = torch.load(os.path.join(lora_dir, "lora_state_dict.pt"))
            for name in lora_state:
                assert name in self.model.state_dict(), f"{name} not in model!"
            self.model.load_state_dict(lora_state, strict=False)

        # Load state
        self.state.load_state(state_path or self.log_dir)

    def finalize(
        self,
    ) -> None:
        """
        Finalizes the logging session.
        """
        # Finalizing `state` synchronizes the state across all processes
        self.state.finalize(log_dir=self.log_dir)

        # Finalizing `logger` flushes and writes the remaining log buffer to disk
        self.logger.finalize()

        # Save configurations
        self.config.save_config(log_dir=self.log_dir)

    def setup(self, log_option_kwargs: Dict[str, Any]) -> None:
        """
        Update logging configurations.

        Args:
            log_option_kwargs: Logging configurations.
        """
        self.logger.opt.setup(log_option_kwargs)

    def save(self, enable: bool = True) -> None:
        """
        Turn on saving.
        """
        self._save = enable

    def eval(self) -> None:
        """
        Set the state of LogIX for testing.
        """
        self.save(False)

    def clear(self) -> None:
        """
        Clear everything in LogIX.
        """
        self.state.clear()
        self.logger.clear()

    def set_logging_config(self, logging_config: LoggingConfig) -> None:
        logging_config.log_dir = self.config.log_dir
        self.config.logging = logging_config

    def set_lora_config(self, lora_config: LoRAConfig) -> None:
        self.config.lora = lora_config

    def get_logging_config(self) -> LoggingConfig:
        return self.config.logging

    def get_lora_config(self) -> LoRAConfig:
        return self.config.lora



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass, field
from typing import List, Optional

import torch.nn as nn


@dataclass
class LogIXArguments:
    project: str = field(
        default="tmp_logix", metadata={"help": "The name of the project."}
    )
    config: str = field(
        default="config.yaml", metadata={"help": "The path to the config file."}
    )
    lora: bool = field(default=False, metadata={"help": "Enable LoRA."})
    hessian: str = field(default="none", metadata={"help": "Hessian type."})
    save: str = field(default="none", metadata={"help": "Save type."})
    initialize_from_log: bool = field(
        default=False, metadata={"help": "Initialize from the log."}
    )
    name_filter: List[str] = field(
        default_factory=list, metadata={"help": "Filter for layer names."}
    )
    type_filter: List[nn.Module] = field(
        default_factory=list, metadata={"help": "Filter for layer types."}
    )
    log_batch_size: int = field(
        default=16, metadata={"help": "The batch size for log dataloader."}
    )
    log_num_workers: int = field(
        default=1, metadata={"help": "The number of workers for log dataloader."}
    )
    input_key: str = field(
        default="input_ids", metadata={"help": "The dictionary key for 'input_ids'."}
    )
    influence_damping: Optional[float] = field(
        default=None, metadata={"help": "A damping term in influence functions."}
    )
    influence_mode: str = field(
        default="dot", metadata={"help": "Influence function mode."}
    )
    influence_groups: Optional[List[str]] = field(
        default=None, metadata={"help": "Influence function groups."}
    )
    label_key: str = field(
        default="labels", metadata={"help": "The dictionary key for 'labels'."}
    )
    attention_key: str = field(
        default="attention_mask",
        metadata={"help": "The dictionary key for 'attention_mask'."},
    )
    data_id: str = field(
        default="detokenize", metadata={"help": "The 'data_id' generation logic."}
    )
    ignore_idx: int = field(
        default=-100, metadata={"help": "The index to be ignored in loss computation."}
    )

    def __post_init__(self):
        self.mode = "log"



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from logix import LogIX
from logix.statistic import CorrectedEigval, Covariance, Log


class LogIXScheduler:
    def __init__(
        self,
        logix: LogIX,
        lora: str = "none",
        hessian: str = "none",
        save: str = "none",
    ):
        self._logix = logix

        self._lora = lora
        self._hessian = hessian
        self._save = save

        self._epoch = -1
        self._logix_state_schedule = []

        self.sanity_check(lora, hessian)
        self.configure_schedule(lora, hessian, save)

        self._schedule_iterator = iter(self._logix_state_schedule)

    def sanity_check(self, lora: str, hessian: str):
        assert lora in ["none", "random", "pca"]
        assert hessian in ["none", "raw", "kfac", "ekfac"]

    def get_lora_epoch(self, lora: str) -> int:
        if lora == "random":
            return 0
        elif lora == "pca":
            return 1
        return -1

    def get_save_epoch(self, save: str) -> int:
        if save != "none":
            return len(self) - 1
        return -1

    def configure_schedule(self, lora: str, hessian: str, save: str) -> None:
        if lora == "pca":
            self._logix_state_schedule.append(
                {"forward": [Covariance], "backward": [Covariance]}
            )

        if hessian == "ekfac":
            self._logix_state_schedule.append(
                {"forward": [Covariance], "backward": [Covariance]}
            )

        last_state = {"forward": [], "backward": [], "grad": []}
        if save != "none":
            last_state[save].append(Log)
        if hessian == "kfac":
            last_state["forward"].append(Covariance)
            last_state["backward"].append(Covariance)
        elif hessian == "ekfac":
            if Log not in last_state["grad"]:
                last_state["grad"].append(Log)
            last_state["grad"].append(CorrectedEigval)
        elif hessian == "raw":
            if Log not in last_state["grad"]:
                last_state["grad"].append(Log)
            last_state["grad"].append(Covariance)
        self._logix_state_schedule.append(last_state)

    def __iter__(self):
        return self

    def __next__(self) -> int:
        logix_state = next(self._schedule_iterator)
        self._epoch += 1

        # maybe add lora
        if self._epoch == self.get_lora_epoch(self._lora):
            self._logix.add_lora()

        # maybe setup save
        if self._epoch == self.get_save_epoch(self._save):
            self._logix.save(True)

        self._logix.setup(logix_state)

        return self._epoch

    def __len__(self) -> int:
        return len(self._logix_state_schedule)



import ast
import re

from setuptools import find_packages, setup


def fetch_requirements(path: str):
    with open(path, "r") as f:
        requirements = f.read().splitlines()
        return requirements


_version_re = re.compile(r"__version__\s+=\s+(.*)")

with open("logix/__init__.py", "rb") as f:
    version = str(
        ast.literal_eval(_version_re.search(f.read().decode("utf-8")).group(1))
    )

python_requires = ">=3.6.0,<3.11.0"

if __name__ == "__main__":
    setup(
        name="logix-ai",
        version=version,
        description="AI Logging for Interpretability and Explainability",
        license="Apache-2.0",
        packages=find_packages(
            exclude=[
                "examples",
                "tests",
                "tests.*",
                "*.tests.*",
            ]
        ),
        install_requires=fetch_requirements("requirements.txt"),
        extras_require={
            "test": fetch_requirements("test-requirements.txt"),
        },
        python_requires=python_requires,
    )



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn
from torch.utils.data import Dataset
from transformers.trainer import *

from logix import LogIX, LogIXScheduler
from logix.huggingface.arguments import LogIXArguments
from logix.huggingface.callback import LogIXCallback
from logix.utils import DataIDGenerator


def patch_trainer(TrainerClass):
    """
    Patch the (variant of) Huggingface Trainer class with the LogIX functionalities.

    Args:
        TrainerClass: The Trainer class to patch
    """

    class PatchedTrainer(TrainerClass):
        def __init__(
            self,
            logix_args: LogIXArguments,
            model: Union[PreTrainedModel, nn.Module] = None,
            args: Optional[TrainingArguments] = None,
            data_collator: Optional[DataCollator] = None,
            train_dataset: Optional[Dataset] = None,
            eval_dataset: Optional[Union[Dataset, Dict[str, Dataset]]] = None,
            tokenizer: Optional[PreTrainedTokenizerBase] = None,
            model_init: Optional[Callable[[], PreTrainedModel]] = None,
            compute_metrics: Optional[Callable[[EvalPrediction], Dict]] = None,
            callbacks: Optional[List[TrainerCallback]] = None,
            optimizers: Tuple[
                torch.optim.Optimizer, torch.optim.lr_scheduler.LambdaLR
            ] = (None, None),
            preprocess_logits_for_metrics: Optional[
                Callable[[torch.Tensor, torch.Tensor], torch.Tensor]
            ] = None,
        ):
            # Initialize LogIX
            self.logix_args = logix_args
            self.logix = LogIX(project=logix_args.project, config=logix_args.config)
            self.logix_scheduler = LogIXScheduler(
                self.logix, hessian=logix_args.hessian, save=logix_args.save
            )
            self.data_id_generator = DataIDGenerator()
            logix_callback = LogIXCallback(
                self.logix, self.logix_scheduler, self.logix_args
            )

            self.logix_input_key = logix_args.input_key
            self.logix_label_key = logix_args.label_key
            self.logix_attention_key = logix_args.attention_key
            self.logix_ignore_idx = logix_args.ignore_idx
            self.data_id_logic = logix_args.data_id

            # Patch TrainingArguments
            if args is None:
                output_dir = "tmp_trainer"
                args = TrainingArguments(output_dir=output_dir)
            args.num_train_epochs = len(self.logix_scheduler)
            args.report_to = []
            args.save_strategy = "no"

            super().__init__(
                model,
                args,
                data_collator,
                train_dataset,
                eval_dataset,
                tokenizer,
                model_init,
                compute_metrics,
                (
                    [logix_callback]
                    if callbacks is None
                    else [logix_callback] + callbacks
                ),
                optimizers,
                preprocess_logits_for_metrics,
            )

        def extract_log(self, *args, **kwargs):
            self.logix_args.mode = "log"
            self.train(*args, **kwargs)

        def influence(self, *args, **kwargs):
            self.logix_args.mode = "influence"
            self.train(*args, **kwargs)

            return self.logix.influence.get_influence_scores()

        def self_influence(self, *args, **kwargs):
            self.logix_args.mode = "self_influence"
            self.train(*args, **kwargs)

            return

        def create_optimizer_and_scheduler(self, num_training_steps: int):
            self.create_optimizer()
            optimizer = self.optimizer
            self.create_scheduler(
                num_training_steps=num_training_steps, optimizer=optimizer
            )

        def create_optimizer(self):
            class DummyOptimizer:
                def __init__(self, params):
                    pass

                def step(self):
                    pass

                def zero_grad(self):
                    pass

                def state_dict(self):
                    return dict()

            self.optimizer = DummyOptimizer(self.model.parameters())

        def create_scheduler(
            self, num_training_steps: int, optimizer: torch.optim.Optimizer = None
        ):
            class DummyScheduler:
                def __init__(self):
                    pass

                def step(self):
                    pass

                def get_last_lr(self):
                    return [0]

                def state_dict(self):
                    return dict()

            self.lr_scheduler = DummyScheduler()
            return self.lr_scheduler

        def training_step(
            self, model: nn.Module, inputs: Dict[str, Union[torch.Tensor, Any]]
        ) -> torch.Tensor:
            model.eval()
            inputs = self._prepare_inputs(inputs)

            data_id = self.get_data_id(inputs)
            mask = inputs.get(self.logix_attention_key, None)
            sum_scale = self.get_sum_scale(inputs)

            with self.logix(data_id=data_id, mask=mask):
                if is_sagemaker_mp_enabled():
                    loss_mb = smp_forward_backward(
                        model, inputs, self.args.gradient_accumulation_steps
                    )
                    return loss_mb.reduce_mean().detach().to(self.args.device)

                with self.compute_loss_context_manager():
                    loss = self.compute_loss(model, inputs)

                if self.args.n_gpu > 1:
                    loss = (
                        loss.mean()
                    )  # mean() to average on multi-gpu parallel training

                # loss reduction with sum instead of mean
                loss = loss * sum_scale
                if self.use_apex:
                    with amp.scale_loss(loss, self.optimizer) as scaled_loss:
                        scaled_loss.backward()
                else:
                    self.accelerator.backward(loss)

            return loss.detach() / self.args.gradient_accumulation_steps

        def get_data_id(self, inputs):
            ipt = inputs.get(self.logix_input_key)
            if self.data_id_logic == "detokenize":
                assert self.tokenizer is not None
                return self.tokenizer.batch_decode(ipt, skip_special_tokens=True)
            elif self.data_id_logic == "hash":
                return self.data_id_generator(ipt)

        def get_sum_scale(self, inputs):
            labels = inputs.get(self.logix_label_key)
            return (labels != self.logix_ignore_idx).sum().item()

    return PatchedTrainer



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from dataclasses import asdict, dataclass, field, is_dataclass
from typing import Any, Dict, List, Optional, Union

import torch
import yaml

from logix.utils import get_rank


def init_config_from_yaml(project: str, logix_config: Optional[str] = None):
    config_dict = {}
    if logix_config is not None:
        assert os.path.exists(logix_config), f"{logix_config} doesn't exist!"
        with open(logix_config, "r", encoding="utf-8") as config_file:
            config_dict = yaml.safe_load(config_file)

    assert project is not None
    config_dict["project"] = project

    return Config(**config_dict)


def load_config_from_dict(config, config_dict: Dict[str, Any]):
    for name, value in config_dict.items():
        if isinstance(value, dict):
            assert is_dataclass(getattr(config, name))
            load_config_from_dict(getattr(config, name), value)
        else:
            setattr(config, name, value)


@dataclass
class LoggingConfig:
    """
    Configuration for logging.

    Args:
        flush_threshold: Flush threshold for the buffer.
        num_workers: Number of workers used for log saving.
        cpu_offload: Offload statistic states to CPU.
    """

    log_dir: str = field(init=False)
    flush_threshold: int = field(
        default=1000000000, metadata={"help": "Flush threshold for the log buffer."}
    )
    num_workers: int = field(
        default=1, metadata={"help": "Number of workers used for logging."}
    )
    cpu_offload: int = field(
        default=False, metadata={"help": "Offload statistic states to CPU."}
    )
    log_dtype: str = field(default="none", metadata={"help": "Data type for logging."})

    def get_dtype(self):
        if self.log_dtype == "none":
            return None
        elif self.log_dtype == "float64":
            return torch.float64
        elif self.log_dtype == "float16":
            return torch.float16
        elif self.log_dtype == "bfloat16":
            return torch.bfloat16
        elif self.log_dtype == "int8":
            return torch.int8
        return torch.float32


@dataclass
class LoRAConfig:
    """
    Configuration for LoRA.

    Args:
        init: Initialization method for LoRA.
        rank: Rank for LoRA.
        parameter_sharing: Parameter sharing for LoRA.
        parameter_sharing_groups: Parameter sharing groups for LoRA.
    """

    init: str = field(
        default="random", metadata={"help": "Initialization method for LoRA."}
    )
    rank: int = field(default=64, metadata={"help": "Rank for LoRA."})
    parameter_sharing: bool = field(
        default=False, metadata={"help": "Parameter sharing for LoRA."}
    )
    parameter_sharing_groups: Optional[List[str]] = field(
        default=None, metadata={"help": "Parameter sharing groups for LoRA."}
    )


@dataclass
class InfluenceConfig:
    """
    Configuration for influence.

    Args:
        damping: Damping for influence.
        relative_damping: Compute the damping term based on sigular values.
        mode: Mode for influence.
    """

    log_dir: str = field(init=False)
    damping: float = field(
        default=1e-10, metadata={"help": "Damping strength for influence."}
    )
    relative_damping: bool = field(
        default=False,
        metadata={"help": "Compute the damping term based on sigular values."},
    )
    mode: str = field(default="dot", metadata={"help": "Mode for influence."})
    flatten: bool = field(
        default=False, metadata={"help": "flattening flag for logging"}
    )


@dataclass
class Config:
    """
    Configuration management class.
    Loads configurations from a YAML file and provides access to component-specific configurations.

    Args:
        project: Project name.
        root_dir: Root directory for logging.
        logging: Logging configuration.
        lora: LoRA configuration.
        analysis: Analysis configuration.
    """

    project: str
    log_dir: str = field(init=False)
    root_dir: str = field(
        default="./logix", metadata={"help": "Root directory for logging."}
    )
    logging: Union[Dict[str, Any], LoggingConfig] = field(
        default_factory=LoggingConfig, metadata={"help": "Logging configuration."}
    )
    lora: Union[Dict[str, Any], LoRAConfig] = field(
        default_factory=LoRAConfig, metadata={"help": "LoRA configuration."}
    )

    def __post_init__(self):
        if isinstance(self.logging, dict):
            self.logging = LoggingConfig(**self.logging)
        if isinstance(self.lora, dict):
            self.lora = LoRAConfig(**self.lora)

        self.log_dir = None
        self.configure_log_dir()

    def configure_log_dir(self) -> None:
        """
        Set single logging directory for all components.
        """
        self.log_dir = os.path.join(self.root_dir, self.project)
        self.logging.log_dir = self.log_dir

        if not os.path.exists(self.log_dir) and get_rank() == 0:
            os.makedirs(self.log_dir)

    def load_config(self, logix_config: str) -> None:
        """
        Load configuration from the saved YAML file.

        Args:
            logix_config: Path to the saved YAML file.
        """
        assert os.path.exists(logix_config), "Configuration file not found."
        config_dict = {}
        with open(logix_config, "r", encoding="utf-8") as config_file:
            config_dict = yaml.safe_load(config_file)

        load_config_from_dict(self, config_dict)
        self.configure_log_dir()

    def save_config(self, log_dir: Optional[str] = None) -> None:
        """
        Save configuration to a YAML file.
        """
        if get_rank() == 0:
            config_file = os.path.join(log_dir or self.log_dir, "config.yaml")
            config_dict = asdict(self)

            with open(config_file, "w", encoding="utf-8") as f:
                yaml.dump(config_dict, f, default_flow_style=False)



# Copyright 2023-present the LogIX team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

from torch.utils.data import Dataset

from logix.logging.log_loader_utils import (
    find_chunk_indices,
    get_flatten_item,
    get_mmap_data,
    get_mmap_metadata,
    unflatten_tensor,
)


class LogDataset(Dataset):
    def __init__(self, log_dir, flatten):
        self.chunk_indices = None
        self.memmaps = []

        self.data_id_to_chunk = OrderedDict()
        self.log_dir = log_dir
        self.flatten = flatten

        # Find all chunk indices
        self.chunk_indices = find_chunk_indices(self.log_dir)
        self.fetch_data()
        self.data_id_list = list(self.data_id_to_chunk.keys())

    def fetch_data(self):
        # Add metadata and mmap files for all indices.
        for idx, chunk_index in enumerate(self.chunk_indices):
            file_root = f"log_{chunk_index}"
            mmap_filename = f"{file_root}.mmap"
            entry = get_mmap_data(self.log_dir, mmap_filename)
            self.memmaps.append(entry)

            self.data_id_to_chunk = get_mmap_metadata(
                self.data_id_to_chunk,
                self.log_dir,
                f"{file_root}_metadata.json",
                idx,
            )

    def __getitem__(self, index):
        data_id = self.data_id_list[index]
        chunk_idx, entry = self.data_id_to_chunk[data_id]
        nested_dict = {}
        mmap = self.memmaps[chunk_idx]
        offset = entry["offset"]
        flat_tensor = get_flatten_item(
            mmap, offset, entry["block_size"], entry["dtype"]
        )
        if self.flatten:
            return data_id, flat_tensor
        start = 0
        for i in range(len(entry["path"])):
            path = entry["path"][i]
            shape = tuple(entry["shape"][i])
            tensor, start = unflatten_tensor(flat_tensor, shape, start)
            current_level = nested_dict
            for key in path[:-1]:
                if key not in current_level:
                    current_level[key] = {}
                current_level = current_level[key]
            current_level[path[-1]] = tensor
        assert (
            entry["block_size"] == start
        ), f"block_size does not match with the shape for data_id: {entry['data_id']}"
        return data_id, nested_dict

    def __len__(self):
        return len(self.data_id_to_chunk)

    def close(self):
        for mmap in self.memmaps:
            del mmap
